from typing import TYPE_CHECKING, Optional

from angr.analyses.decompiler.decompilation_options import options as dec_options
from angr.analyses.decompiler.optimization_passes import get_default_optimization_passes, get_optimization_passes
from angr.analyses.decompiler.peephole_optimizations import EXPR_OPTS, STMT_OPTS
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QComboBox, QLineEdit, QPushButton, QTreeWidget, QTreeWidgetItem, QVBoxLayout, QWidget

if TYPE_CHECKING:
    from angrmanagement.ui.views.code_view import CodeView


class OptionType:
    """
    An enum to determine what the .option field of a QDecompilationOption contains
    """

    OPTION = 1
    OPTIMIZATION_PASS = 2
    PEEPHOLE_OPTIMIZATION = 3


class QDecompilationOption(QTreeWidgetItem):
    """
    The UI entry for a single decompliation option. Get status with item.state.
    """

    def __init__(self, parent, option, type_: int, enabled=True):
        super().__init__(parent)
        self.option = option
        self.type = type_

        # optional and may not exist
        self._combo_box = None

        if self.type == OptionType.OPTIMIZATION_PASS:
            self.setText(0, option.NAME)
            self.setToolTip(0, option.DESCRIPTION)
        elif self.type == OptionType.OPTION:
            self.setText(0, option.NAME)
            self.setToolTip(0, option.DESCRIPTION)
        elif self.type == OptionType.PEEPHOLE_OPTIMIZATION:
            self.setText(0, option.NAME)
            self.setToolTip(0, option.DESCRIPTION)
        else:
            raise NotImplementedError("Unsupported option type %s." % self.type_)

        # should make a dropdown option
        if hasattr(self.option, "value_type") and self.option.value_type != bool and self.option.candidate_values:
            self._combo_box = QComboBox()
            self._combo_box.addItems(
                [self.option.default_value]
                + [c for c in self.option.candidate_values if c != self.option.default_value]
            )
            self._combo_box.setToolTip(f"{option.NAME}: {option.DESCRIPTION}")
            # XXX: causes an itemChanged event for the tree
            self._combo_box.currentTextChanged.connect(lambda x: self.setText(0, self._combo_box.currentText()))
            self.treeWidget().setItemWidget(self, 0, self._combo_box)

        # should make a boolean click option
        else:
            self.setFlags(self.flags() | Qt.ItemIsUserCheckable)
            if enabled:
                self.setCheckState(0, Qt.Checked)
            else:
                self.setCheckState(0, Qt.Unchecked)

    @property
    def state(self):
        if self._combo_box:
            return self._combo_box.currentText()
        else:
            return bool(self.checkState(0) == Qt.CheckState.Checked)


class QDecompilationOptions(QWidget):
    """
    The widget for selecting values for decompilation options. Will synchronize its status back to its parent (passed
    in parameter) code view with the Apply button is pressed.

    Since some options have default values depending on the current arch and os, it is important to call
    reload(force=True) to reset values to their defaults whenever the current project changes.
    """

    def __init__(self, code_view, instance):
        super().__init__()

        self.dirty = True

        self._code_view: CodeView = code_view
        self._instance = instance
        self._options = None
        self._opti_passes = None
        self._peephole_opts = None

        # widgets
        self._search_box: QLineEdit
        self._treewidget: QTreeWidget
        self._apply_btn: QPushButton

        self._qoptions = []
        self._qoptipasses = []
        self._qpeephole_opts = []

        self._init_widgets()

        self.reload(True)

    def reload(self, force=False):
        if force or self._options is None:
            self._options = self.get_default_options()

        if force or self._opti_passes is None:
            if not self._instance.project.am_none:
                self._opti_passes = self.get_all_passes()
            else:
                self._opti_passes = []

        if force or self._peephole_opts is None:
            self._peephole_opts = self.get_all_peephole_opts()

        self._reload_options(force)
        self._set_visibility(self._search_box.text())

    def _on_item_changed(self, item, _column):
        if getattr(item.option, "clears_cache", True):
            self.dirty = True

    def _on_apply_pressed(self):
        if self.dirty:
            self.dirty = False
            # clear the cached version
            self._code_view.decompile(reset_cache=True)
        else:
            if not self._code_view.codegen.am_none:
                self._code_view.codegen.reapply_options(self.option_and_values)
                self._code_view.codegen.am_event()

    @property
    def selected_passes(self):
        selected = []
        for item in self._qoptipasses:
            if item.state:
                selected.append(item.option)
        return selected

    @property
    def selected_peephole_opts(self):
        selected = []
        for item in self._qpeephole_opts:
            if item.state:
                selected.append(item.option)
        return selected

    @property
    def option_and_values(self):
        return [(item.option, item.state) for item in self._qoptions]

    def get_default_options(self):  # pylint: disable=no-self-use
        return dec_options

    def get_default_passes(self):
        if self._instance is None or self._instance.project.am_none:
            return []
        return get_default_optimization_passes(self._instance.project.arch, self._instance.project.simos.name) + [
            x for x, de, in self._code_view.workspace.plugins.optimization_passes() if de
        ]

    def get_all_passes(self):
        if self._instance is None or self._instance.project.am_none:
            return []
        return get_optimization_passes(self._instance.project.arch, self._instance.project.simos.name) + [
            x for x, _, in self._code_view.workspace.plugins.optimization_passes()
        ]

    def get_default_peephole_opts(self):  # pylint: disable=no-self-use
        return STMT_OPTS + EXPR_OPTS

    def get_all_peephole_opts(self):  # pylint: disable=no-self-use
        return STMT_OPTS + EXPR_OPTS

    def _init_widgets(self):
        # search box
        self._search_box = QLineEdit()
        self._search_box.textChanged.connect(self._on_search_box_text_changed)

        # tree view
        self._treewidget = QTreeWidget()
        self._treewidget.setHeaderHidden(True)
        self._treewidget.itemChanged.connect(self._on_item_changed)

        # refresh button
        self._apply_btn = QPushButton("Apply")
        self._apply_btn.clicked.connect(self._on_apply_pressed)

        layout = QVBoxLayout()
        layout.addWidget(self._search_box)
        layout.addWidget(self._treewidget)
        layout.addWidget(self._apply_btn)

        self.setLayout(layout)

    def _reload_options(self, reset_values=False):
        vals_options = dict(self.option_and_values)
        vals_peephole = self.selected_peephole_opts
        vals_passes = self.selected_passes

        self._treewidget.clear()
        self._qoptions.clear()
        self._qoptipasses.clear()
        self._qpeephole_opts.clear()

        categories = {}

        # populate the tree widget with new options
        for option in sorted(self._options, key=lambda x: x.NAME):
            if option.category in categories:
                category = categories[option.category]
            else:
                category = QTreeWidgetItem(self._treewidget, [option.category])
                categories[option.category] = category

            enabled = option.default_value if reset_values else vals_options.get(option, option.default_value)
            w = QDecompilationOption(category, option, OptionType.OPTION, enabled=enabled)
            self._qoptions.append(w)

        passes_category = QTreeWidgetItem(self._treewidget, ["Optimization Passes"])
        categories["passes"] = passes_category

        default_passes = set(self.get_default_passes())
        for pass_ in self._opti_passes:
            enabled = pass_ in default_passes if reset_values else pass_ in vals_passes
            w = QDecompilationOption(passes_category, pass_, OptionType.OPTIMIZATION_PASS, enabled=enabled)
            self._qoptipasses.append(w)

        po_category = QTreeWidgetItem(self._treewidget, ["Peephole Optimizations"])
        categories["peephole_opts"] = po_category

        default_peephole_opts = self.get_default_peephole_opts()
        for opt_ in self._peephole_opts:
            enabled = opt_ in default_peephole_opts if reset_values else opt_ in vals_peephole
            w = QDecompilationOption(po_category, opt_, OptionType.PEEPHOLE_OPTIMIZATION, enabled=enabled)
            self._qpeephole_opts.append(w)

        # expand all
        self._treewidget.expandAll()

    def _set_visibility(self, filter_by: Optional[str] = None):
        for w in self._qoptions:
            w.setHidden(
                bool(filter_by) and not (filter_by in w.option.NAME.lower() or filter_by in w.option.category.lower())
            )
        for w in self._qoptipasses:
            w.setHidden(
                bool(filter_by) and not (filter_by in w.option.__name__.lower() or filter_by in w.option.NAME.lower())
            )
        for w in self._qpeephole_opts:
            w.setHidden(
                bool(filter_by)
                and not (filter_by in w.option.NAME.lower() or filter_by in w.option.DESCRIPTION.lower())
            )

    def _on_search_box_text_changed(self, text: str):
        self._set_visibility(filter_by=text)



class IndirectJump:
    """
    A meta class describing an indirect jump instruction.
    """

    def __init__(self, mnemonic, resolved, targets=None):
        self.mnemonic = mnemonic
        self.resolved = resolved
        self.targets = targets



from typing import TYPE_CHECKING, Optional

from angr.analyses.decompiler.clinic import Clinic
from angr.analyses.disassembly import Instruction, IROp
from angr.sim_variable import SimRegisterVariable
from PySide6.QtCore import QMarginsF, QRectF
from PySide6.QtGui import QPainterPath, QPen

from angrmanagement.config import Conf
from angrmanagement.utils import get_block_objects, get_label_text, get_out_branches_for_insn
from angrmanagement.utils.block_objects import FunctionHeader, Label, PhiVariable, Variables

from .qblock_code import QAilObj, QBlockCode, QBlockCodeOptions, QIROpObj
from .qblock_label import QBlockLabel
from .qfunction_header import QFunctionHeader
from .qgraph_object import QCachedGraphicsItem
from .qinstruction import QInstruction
from .qphivariable import QPhiVariable
from .qvariable import QVariable

if TYPE_CHECKING:
    from PySide6.QtWidgets import QGraphicsPathItem


class QBlock(QCachedGraphicsItem):
    TOP_PADDING = 5
    BOTTOM_PADDING = 5
    LEFT_PADDING = 10
    RIGHT_PADDING = 10
    SPACING = 0
    AIL_SHOW_CONDITIONAL_JUMP_TARGETS = True
    SHADOW_OFFSET_X = 0
    SHADOW_OFFSET_Y = 0

    def __init__(
        self,
        instance,
        func_addr,
        disasm_view,
        disasm,
        infodock,
        addr,
        cfg_nodes,
        out_branches,
        scene,
        parent=None,
        idx=None,
    ):
        super().__init__(parent=parent)

        # initialization
        self.instance = instance
        self.func_addr = func_addr
        self.disasm_view = disasm_view
        self.disasm = disasm
        self.infodock = infodock
        self.variable_manager = infodock.variable_manager
        self.addr = addr
        self.cfg_nodes = cfg_nodes
        self.out_branches = out_branches
        self.scene = scene
        self.idx = idx

        self._config = Conf

        self.objects = []  # instructions and labels
        self._block_item: Optional[QPainterPath] = None
        self._block_item_obj: Optional[QGraphicsPathItem] = None
        self.addr_to_insns = {}
        self.addr_to_labels = {}
        self.qblock_annotations = {}

        self._block_code_options: QBlockCodeOptions = QBlockCodeOptions()
        self._update_block_code_options()

        self._init_widgets()

        self._objects_are_hidden = False
        self._objects_are_temporarily_hidden = False

        self._create_block_item()

        self.setAcceptHoverEvents(True)

    #
    # Properties
    #

    @property
    def mode(self):
        raise NotImplementedError

    @property
    def width(self):
        return self.boundingRect().width()

    @property
    def height(self):
        return self.boundingRect().height()

    #
    # Public methods
    #

    def clear_cache(self):
        super().clear_cache()
        for obj in self.objects:
            obj.clear_cache()

    def _update_block_code_options(self):
        self._block_code_options.show_conditional_jump_targets = self.AIL_SHOW_CONDITIONAL_JUMP_TARGETS
        self._block_code_options.show_variables = self.disasm_view.show_variable
        self._block_code_options.show_variable_identifiers = self.disasm_view.show_variable_identifier

    def refresh(self):
        self._update_block_code_options()
        for obj in self.objects:
            obj.refresh()
        self.layout_widgets()
        self.recalculate_size()
        self._create_block_item()
        self.update()

    def reload(self):
        self._init_widgets()
        self.refresh()

    def size(self):
        return self.width, self.height

    def instruction_position(self, insn_addr):
        if insn_addr in self.addr_to_insns:
            insn = self.addr_to_insns[insn_addr]
            pos = insn.pos()
            return pos.x(), pos.y()

        return None

    #
    # Initialization
    #

    def _create_block_item(self):
        """
        Create the block background and border.
        """
        if self._block_item_obj is not None and self.scene is not None:
            self.scene.removeItem(self._block_item_obj)
            self._block_item = None
            self._block_item_obj = None

        self._block_item = QPainterPath()
        self._block_item.addRoundedRect(
            0,
            0,
            self.width - self.SHADOW_OFFSET_X,
            self.height - self.SHADOW_OFFSET_Y,
            self._config.disasm_view_node_rounding,
            self._config.disasm_view_node_rounding,
        )

    def _init_ail_block_widgets(self):
        bn = self.cfg_nodes
        if bn.addr in self.disasm.kb.labels:
            label = QBlockLabel(
                bn.addr,
                get_label_text(bn.addr, self.disasm.kb),
                self._config,
                self.disasm_view,
                self.instance,
                self.infodock,
                parent=self,
            )
            self.objects.append(label)
            self.addr_to_labels[bn.addr] = label

        # always add the block name as a label and instruction:
        block_name_label = QBlockLabel(
            bn.addr, f"loc_{hex(bn.addr)}:", self._config, self.disasm_view, self.instance, self.infodock, parent=self
        )
        self.objects.append(block_name_label)
        self.addr_to_labels[bn.addr] = block_name_label

        for stmt in bn.statements:
            code_obj = QAilObj(stmt, self.instance, self.infodock, parent=None, options=self._block_code_options)
            obj = QBlockCode(
                stmt.ins_addr, code_obj, self._config, self.disasm_view, self.instance, self.infodock, parent=self
            )
            code_obj.parent = obj  # Reparent
            self.objects.append(obj)
            self.addr_to_insns[bn.addr] = obj

    def _init_disassembly_block_widgets(self):
        for obj in get_block_objects(self.disasm, self.cfg_nodes, self.func_addr):
            if isinstance(obj, Instruction):
                out_branch = get_out_branches_for_insn(self.out_branches, obj.addr)
                insn = QInstruction(
                    self.instance,
                    self.func_addr,
                    self.disasm_view,
                    self.disasm,
                    self.infodock,
                    obj,
                    out_branch,
                    self._config,
                    parent=self,
                )
                self.objects.append(insn)
                self.addr_to_insns[obj.addr] = insn
            elif isinstance(obj, Label):
                label = QBlockLabel(
                    obj.addr, obj.text, self._config, self.disasm_view, self.instance, self.infodock, parent=self
                )
                self.objects.append(label)
                self.addr_to_labels[obj.addr] = label
            elif isinstance(obj, IROp):
                code_obj = QIROpObj(obj, self.infodock, parent=None)
                disp_obj = QBlockCode(
                    obj.addr, code_obj, self._config, self.disasm_view, self.instance, self.infodock, parent=self
                )
                code_obj.parent = disp_obj  # Reparent
                self.objects.append(disp_obj)
            elif isinstance(obj, PhiVariable):
                if not isinstance(obj.variable, SimRegisterVariable):
                    phivariable = QPhiVariable(self.instance, self.disasm_view, obj, self._config, parent=self)
                    self.objects.append(phivariable)
            elif isinstance(obj, Variables):
                for var in obj.variables:
                    variable = QVariable(self.instance, self.disasm_view, var, self._config, self.infodock, parent=self)
                    self.objects.append(variable)
            elif isinstance(obj, FunctionHeader):
                self.objects.append(
                    QFunctionHeader(
                        self.func_addr,
                        obj.name,
                        obj.prototype,
                        obj.args,
                        self._config,
                        self.disasm_view,
                        self.instance,
                        self.infodock,
                        parent=self,
                    )
                )

    def _init_widgets(self):
        if self.scene is not None:
            for obj in self.objects:
                self.scene.removeItem(obj)

        self.objects.clear()

        if isinstance(self.disasm, Clinic):
            self._init_ail_block_widgets()
        else:
            self._init_disassembly_block_widgets()

        self.layout_widgets()

    def layout_widgets(self):
        raise NotImplementedError


class QGraphBlock(QBlock):
    MINIMUM_DETAIL_LEVEL = 0.4
    AIL_SHOW_CONDITIONAL_JUMP_TARGETS = False
    SHADOW_OFFSET_X = 5
    SHADOW_OFFSET_Y = 5
    BLOCK_ANNOTATIONS_LEFT_PADDING = 2

    @property
    def mode(self):
        return "graph"

    def layout_widgets(self):
        x, y = self.LEFT_PADDING, self.TOP_PADDING

        if self.qblock_annotations and self.qblock_annotations.scene():
            self.qblock_annotations.scene().removeItem(self.qblock_annotations)

        self.qblock_annotations = self.disasm_view.fetch_qblock_annotations(self)

        for obj in self.objects:
            if self.qblock_annotations.width > 0:
                obj.setPos(self.BLOCK_ANNOTATIONS_LEFT_PADDING + self.qblock_annotations.width + x, y)
            else:
                obj.setPos(x, y)
            if isinstance(obj, QInstruction) and self.qblock_annotations.get(obj.addr):
                qinsn_annotations = self.qblock_annotations.get(obj.addr)
                for qinsn_annotation in qinsn_annotations:
                    qinsn_annotation.setY(obj.y())
            y += obj.boundingRect().height()

    def hoverEnterEvent(self, event):
        self.infodock.hover_block(self.addr)
        event.accept()

    def hoverLeaveEvent(self, event):
        self.infodock.unhover_block(self.addr)
        event.accept()

    def mousePressEvent(self, event):
        if self.disasm_view.workspace.plugins.handle_click_block(self, event):
            # stop handling this event if the event has been handled by a plugin
            event.accept()
            return

        # the block is selected
        self.on_selected()

        super().mousePressEvent(event)

    def _calc_backcolor(self, should_omit_text):
        color = self.disasm_view.workspace.plugins.color_block(self.addr)
        if color is not None:
            return color

        if should_omit_text:
            return self._config.disasm_view_node_zoomed_out_background_color

        return self._config.disasm_view_node_background_color

    def _set_block_objects_visibility(self, visible: bool):
        for obj in self.objects:
            obj.setVisible(visible)
            obj.setEnabled(visible)

    def restore_temporarily_hidden_objects(self):
        if self._objects_are_temporarily_hidden != self._objects_are_hidden:
            self._set_block_objects_visibility(not self._objects_are_hidden)
            self._objects_are_temporarily_hidden = self._objects_are_hidden

    def paint(self, painter, option, widget):  # pylint: disable=unused-argument
        lod = option.levelOfDetailFromTransform(painter.worldTransform())
        should_omit_text = lod < QGraphBlock.MINIMUM_DETAIL_LEVEL

        painter.setBrush(self._config.disasm_view_node_shadow_color)
        painter.setPen(self._config.disasm_view_node_shadow_color)
        shadow_path = QPainterPath(self._block_item)
        shadow_path.translate(self.SHADOW_OFFSET_X, self.SHADOW_OFFSET_Y)
        painter.drawPath(shadow_path)

        # background of the node
        painter.setBrush(self._calc_backcolor(should_omit_text))
        if self.infodock.is_block_selected(self.addr):
            painter.setPen(QPen(self._config.disasm_view_selected_node_border_color, 2.5))
        else:
            painter.setPen(QPen(self._config.disasm_view_node_border_color, 1.5))

        self._block_item_obj = painter.drawPath(self._block_item)

        # content drawing is handled by qt since children are actual child widgets

        # if we are too far zoomed out, do not draw the text
        if self._objects_are_hidden != should_omit_text:
            self._set_block_objects_visibility(not should_omit_text)
            view = self.scene.parent()
            if view.is_extra_render_pass:
                self._objects_are_temporarily_hidden = should_omit_text
            else:
                self._objects_are_hidden = should_omit_text

        # extra content
        self.disasm_view.workspace.plugins.draw_block(self, painter)

    def on_selected(self):
        self.infodock.select_block(self.addr)

    def _boundingRect(self):
        cbr = self.childrenBoundingRect()
        margins = QMarginsF(
            self.LEFT_PADDING,
            self.TOP_PADDING,
            self.RIGHT_PADDING + self.SHADOW_OFFSET_X,
            self.BOTTOM_PADDING + self.SHADOW_OFFSET_Y,
        )
        return cbr.marginsAdded(margins)


class QLinearBlock(QBlock):
    ADDRESS_PADDING = 10

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._height = 0
        self._width = 0

    @property
    def mode(self):
        return "linear"

    @staticmethod
    def format_address(addr):
        return f"{addr:08x}"

    def layout_widgets(self):
        y_offset = 0

        max_width = 0

        for obj in self.objects:
            y_offset += self.SPACING
            obj_start = 0
            obj.setPos(obj_start, y_offset)
            if obj_start + obj.width > max_width:
                max_width = obj_start + obj.boundingRect().width()
            y_offset += obj.boundingRect().height()

        self._height = y_offset
        self._width = max_width

    def paint(self, painter, option, widget):  # pylint: disable=unused-argument
        painter.setFont(self._config.disasm_font)

    def _boundingRect(self):
        return QRectF(0, 0, self._width, self._height)



import re
from typing import TYPE_CHECKING

from angr.knowledge_plugins import Function
from PySide6.QtCore import QSize
from PySide6.QtWidgets import QCheckBox, QHBoxLayout, QLabel, QLineEdit, QVBoxLayout

from angrmanagement.ui.widgets.qfunction_combobox import QFunctionComboBox
from angrmanagement.ui.widgets.qstring_table import QStringTable

from .view import BaseView

if TYPE_CHECKING:
    from angr.knowledge_plugins.cfg.memory_data import MemoryData


class StringsView(BaseView):
    def __init__(self, workspace, instance, default_docking_position, *args, **kwargs):
        super().__init__("strings", workspace, instance, default_docking_position, *args, **kwargs)

        self.base_caption = "Strings"

        self._string_table: QStringTable
        self._function_list: QFunctionComboBox

        self._selected_function = None

        self._init_widgets()
        self.reload()

    def reload(self):
        if self.instance.kb is None:
            return
        self._function_list.functions = self.instance.kb.functions
        self._string_table.cfg = self.instance.cfg
        self._string_table.xrefs = self.instance.project.kb.xrefs
        self._string_table.function = self._selected_function

    def sizeHint(self):
        return QSize(400, 800)

    #
    # Event handlers
    #

    def _on_function_selected(self, function):
        if isinstance(function, str) and function == "all":
            # all functions
            self._selected_function = None

        elif isinstance(function, Function):
            self._selected_function = function

        self.reload()

    def _on_string_selected(self, s: "MemoryData"):
        """
        A string reference is selected.

        :param s:
        :return:
        """

        if len(self.workspace.view_manager.views_by_category["disassembly"]) == 1:
            disasm_view = self.workspace.view_manager.first_view_in_category("disassembly")
        else:
            disasm_view = self.workspace.view_manager.current_view_in_category("disassembly")
        if disasm_view is not None:
            disasm_view.jump_to(s.addr)
            disasm_view.select_label(s.addr)
            self.workspace.view_manager.raise_view(disasm_view)

    def on_filter_change(self, **kwargs):  # pylint: disable=unused-argument
        pattern = self._filter_string.text()
        regex = self._regex_checkbox.isChecked()
        if regex:
            try:
                pattern = re.compile(pattern)
            except re.error:
                return
        self._string_table.filter_string = pattern

    #
    # Private methods
    #

    def _init_widgets(self):
        self._function_list = QFunctionComboBox(
            show_all_functions=True, selection_callback=self._on_function_selected, parent=self
        )

        self._filter_string = QLineEdit(self)
        self._regex_checkbox = QCheckBox("Regex", self)
        self._filter_string.textChanged.connect(self.on_filter_change)
        self._regex_checkbox.stateChanged.connect(self.on_filter_change)

        function_layout = QHBoxLayout()
        function_layout.addWidget(QLabel("Function:", self))
        function_layout.addWidget(self._function_list, 10)
        function_layout.addWidget(QLabel("Filter:", self))
        function_layout.addWidget(self._filter_string, 10)
        function_layout.addWidget(self._regex_checkbox)
        function_layout.setContentsMargins(3, 3, 3, 3)
        function_layout.setSpacing(3)

        self._string_table = QStringTable(self.instance, self, selection_callback=self._on_string_selected)

        layout = QVBoxLayout()
        layout.addLayout(function_layout)
        layout.addWidget(self._string_table)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)



from PySide6.QtWidgets import QVBoxLayout

from angrmanagement.ui.widgets.qfunction_table import QFunctionTable

from .view import BaseView


class FunctionsView(BaseView):
    """
    View displaying functions in the project.
    """

    def __init__(self, workspace, instance, default_docking_position, *args, **kwargs):
        super().__init__("functions", workspace, instance, default_docking_position, *args, **kwargs)

        self.base_caption = "Functions"
        self._function_table: QFunctionTable

        self.instance.cfg.am_subscribe(self.reload)

        self._init_widgets()

        self.width_hint = 300
        self.height_hint = 0
        self.updateGeometry()

        self.function_count = None
        self._displayed_function_count = None

        self.reload()

    #
    # Public methods
    #

    def refresh(self):
        self._function_table.refresh()

    def set_function_count(self, count):
        self.function_count = count

    def set_displayed_function_count(self, count):
        self._displayed_function_count = count

    def reload(self):
        if not self.instance.cfg.am_none:
            self._function_table.function_manager = self.instance.kb.functions

    def subscribe_func_select(self, callback):
        """
        Appends the provided function to the list of callbacks to be called when a function is selected in the
        functions table. The callback's only parameter is the `angr.knowledge_plugins.functions.function.Function`
        :param callback: The callback function to call, which must accept **kwargs
        """
        self._function_table.subscribe_func_select(callback)

    #
    # Private methods
    #

    def _init_widgets(self):
        self._function_table = QFunctionTable(self, self.instance, selection_callback=self._on_function_selected)

        vlayout = QVBoxLayout()
        vlayout.addWidget(self._function_table)
        vlayout.setSpacing(0)
        vlayout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(vlayout)

    def _on_function_selected(self, func):
        """
        A new function is on selection right now. Update the disassembly view that is currently at front.

        :param function:
        :return:
        """
        self.workspace.on_function_selected(func=func)



from typing import TYPE_CHECKING, List, Mapping, Optional, Sequence

from PySide6.QtCore import QSize
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QFrame, QMenu

if TYPE_CHECKING:
    import PySide6.QtGui

    from angrmanagement.data.highlight_region import SynchronizedHighlightRegion
    from angrmanagement.ui.workspace import Instance, Workspace


class BaseView(QFrame):
    """
    Base class for all main views.
    """

    # if this view has function-specific views to show. function setter must be implemented when this property is set
    # to True.
    FUNCTION_SPECIFIC_VIEW = False

    def __init__(
        self,
        category: str,
        workspace: "Workspace",
        instance: "Instance",
        default_docking_position: str,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.workspace = workspace
        self.instance: Instance = instance
        self.category = category
        self.default_docking_position = default_docking_position

        self.old_width = None
        self.old_height = None
        self.width_hint = -1
        self.height_hint = -1
        self.index: int = 1
        self.base_caption: str = "View"

    @property
    def function(self):
        raise NotImplementedError

    @function.setter
    def function(self, v):
        raise NotImplementedError

    def is_shown(self):
        return self.visibleRegion().isEmpty() is False

    def focus(self):
        self.workspace.view_manager.raise_view(self)

    def refresh(self):
        pass

    def reload(self):
        pass

    def sizeHint(self):
        return QSize(self.width_hint, self.height_hint)

    def resizeEvent(self, event):
        # Update current width
        self.old_width = event.oldSize().width()
        self.old_height = event.oldSize().height()

    def closeEvent(self, event: "PySide6.QtGui.QCloseEvent"):
        self.workspace.view_manager.remove_view(self)
        event.accept()

    def mainWindowInitializedEvent(self):
        pass

    #
    # Properties
    #

    @property
    def caption(self):
        s = self.base_caption
        if self.index > 1:
            s += f"-{self.index}"
        return s


class ViewState:
    """
    A basic view state to be published through ViewStatePublisherMixin.
    """

    def __init__(self, cursors: Optional[List[int]] = None):
        self.cursors: List[int] = cursors or []


class ViewStatePublisherMixin:
    """
    Views that wish to update the instance 'active' view state for common visualization use this mixin.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.published_view_state = ViewState()

    def on_focused(self):
        self.notify_view_state_updated()

    def notify_view_state_updated(self):
        if self.workspace.view_manager.most_recently_focused_view is self:
            self.instance.active_view_state.am_obj = self.published_view_state
            self.instance.active_view_state.am_event()


class SynchronizedViewState:
    """
    Simple state tracking for synchronized views.
    """

    def __init__(self):
        self.views = set()
        self.cursor_address: Optional[int] = None
        self.highlight_regions: Mapping[SynchronizedView, Sequence[SynchronizedHighlightRegion]] = {}

    def register_view(self, view: "SynchronizedView"):
        """
        Register a synchronized view.
        """
        self.views.add(view)
        for v in self.views:
            v.on_synchronized_view_group_changed()

    def unregister_view(self, view: "SynchronizedView"):
        """
        Unregister a synchronized view.
        """
        rgns = self.highlight_regions.pop(view, None)
        if rgns is not None:
            for v in self.views:
                v.on_synchronized_highlight_regions_changed()
        self.views.remove(view)
        for v in self.views:
            v.on_synchronized_view_group_changed()


class SynchronizedView(BaseView):
    """
    Base class for views which can be synchronized.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._processing_synchronized_cursor_update: bool = False
        self.sync_state: SynchronizedViewState = SynchronizedViewState()
        self.sync_state.register_view(self)

    def desync(self):
        """
        Stop synchronization with any previously synchronized views.
        """
        self.sync_with_state_object()

    def sync_with_state_object(self, state: Optional[SynchronizedViewState] = None):
        """
        Synchronize with another view state.
        """
        self.sync_state.unregister_view(self)
        self.sync_state = state or SynchronizedViewState()
        self.sync_state.register_view(self)
        self.sync_from_state()

    def sync_with_view(self, view: "SynchronizedView"):
        """
        Synchronize with another view.
        """
        self.sync_with_state_object(view.sync_state)

    def sync_from_state(self):
        """
        Update this view to reflect the synchronized view state.
        """
        self.on_synchronized_cursor_address_changed()
        self.on_synchronized_highlight_regions_changed()

    def set_synchronized_cursor_address(self, address: Optional[int]):
        """
        Set synchronized cursor address.
        """
        if not self._processing_synchronized_cursor_update:
            self.sync_state.cursor_address = address
            for view in self.sync_state.views:
                if view is not self:
                    view.on_synchronized_cursor_address_changed()

    def on_synchronized_cursor_address_changed(self):
        """
        Handle synchronized cursor address change event.
        """
        assert not self._processing_synchronized_cursor_update
        self._processing_synchronized_cursor_update = True
        try:
            if self.sync_state.cursor_address is not None:
                self.jump_to(self.sync_state.cursor_address)
        finally:
            self._processing_synchronized_cursor_update = False

    def set_synchronized_highlight_regions(self, regions: Sequence["SynchronizedHighlightRegion"]):
        """
        Set synchronized highlight regions for this view.
        """
        self.sync_state.highlight_regions[self] = regions
        for view in self.sync_state.views:
            if view is not self:
                view.on_synchronized_highlight_regions_changed()

    def on_synchronized_highlight_regions_changed(self):
        """
        Handle synchronized highlight region change event.
        """

    def on_synchronized_view_group_changed(self):
        """
        Handle view being added to or removed from the view synchronization group.
        """

    def closeEvent(self, event: "PySide6.QtGui.QCloseEvent"):
        """
        View close event handler.
        """
        self.desync()
        super().closeEvent(event)

    def get_synchronize_with_submenu(self) -> QMenu:
        """
        Get submenu for 'Synchronize with' context menu.
        """
        mnu = QMenu("&Synchronize with", self)
        groups = {
            v.sync_state
            for v in self.workspace.view_manager.views
            if (v is not self) and isinstance(v, SynchronizedView)
        }
        if len(groups) == 0:
            act = QAction("None available", self)
            act.setEnabled(False)
            mnu.addAction(act)
        else:
            for group in groups:
                act = QAction(", ".join([v.caption for v in group.views if v is not self]), self)
                act.setCheckable(True)
                act.setChecked(group is self.sync_state)
                act.toggled.connect(lambda checked, s=group: self.sync_with_state_object(s if checked else None))
                mnu.addAction(act)
        return mnu



from typing import TYPE_CHECKING, Optional

from PySide6.QtWidgets import QComboBox

if TYPE_CHECKING:
    from angr.knowledge_plugins import Function, FunctionManager


class QFunctionComboBox(QComboBox):
    def __init__(self, show_all_functions=False, selection_callback=None, parent=None):
        super().__init__(parent)

        self._show_all_functions = show_all_functions
        self._selection_callback = selection_callback

        self._function_manager: Optional[FunctionManager] = None

        self.currentIndexChanged.connect(self._on_current_index_changed)

    #
    # Properties
    #

    @property
    def functions(self):
        return self._function_manager

    @functions.setter
    def functions(self, v):
        if v is not self._function_manager:
            self._function_manager = v
            self.reload()

    #
    # Public methods
    #

    def reload(self):
        if self._function_manager is None:
            return

        self.clear()

        if self._show_all_functions:
            self.addItem("All functions", "all")

        for function in self._function_manager.values():
            self.addItem(self._repr_function(function), function)

    #
    # Event handlers
    #

    def _on_current_index_changed(self):
        idx = self.currentIndex()
        if idx == -1:
            return

        function = self.itemData(idx)

        self._selection_callback(function)

    #
    # Private functions
    #

    @staticmethod
    def _repr_function(func: "Function") -> str:
        demangled_name = func.demangled_name
        if len(demangled_name) > 30:
            demangled_name = demangled_name[:30] + "..."
        return f"{demangled_name} ({func.addr:#x})"



import ailment

from angrmanagement.plugins import BasePlugin

from .asm_output import AsmOutput


class AIL2ARM32(BasePlugin):
    """
    Assemble expressions into assembly.
    """

    def __init__(self, workspace, sp_adjust=0x30):
        super().__init__(workspace)
        self.sp_adjust = sp_adjust
        self.registers_in_use = []
        self.free_registers = ["r4", "r5", "r6", "r7", "r8", "r9", "r10", "r11"]
        self.free_registers.reverse()

    def reserve_register(self):
        if not self.free_registers:
            raise Exception("No more registers available")
        reg = self.free_registers.pop()
        self.registers_in_use.append(reg)
        return reg

    def return_register(self, reg):
        self.registers_in_use.remove(reg)
        self.free_registers.append(reg)

    @staticmethod
    def opcode_variant(opcode, size=None, signed=None, cond=None):
        if size is not None:
            if size == 4:
                pass
            elif size == 2:
                if signed is not None and signed:
                    opcode += "s"
                opcode += "h"
            elif size == 1:
                if signed is not None and signed:
                    opcode += "s"
                opcode += "b"
            else:
                raise Exception(f"unsupported size {size} for opcode {opcode}")
        if cond is not None:
            opcode += cond
        return opcode

    def bp_offset_to_sp_offset(self, offset_from_bp, function_addr, expr_addr, adjust=0):
        if function_addr is None or expr_addr is None:
            raise Exception("function_addr and expr_addr must be specified")
        sp = self.workspace.main_instance.project.arch.sp_offset
        bp = self.workspace.main_instance.project.arch.bp_offset
        sptracker = self.workspace.main_instance.project.analyses.StackPointerTracker(
            self.workspace.main_instance.project.kb.functions[function_addr], {sp, bp}
        )
        sp_offset = sptracker.offset_after(expr_addr, sp)
        bp_offset = sptracker.offset_after(expr_addr, bp)
        if sp_offset >= (1 << (32 - 1)):
            sp_offset -= 1 << 32
        if bp_offset >= (1 << (32 - 1)):
            bp_offset -= 1 << 32
        return bp_offset + offset_from_bp - sp_offset + adjust

    def assemble(self, expr, function_addr=None, expr_addr=None):
        dest, asm, const_pool = self.assemble_recursive(expr, function_addr, expr_addr)
        self.return_register(dest)
        asm += f"mov r0, {dest}\n"
        if const_pool:
            asm += "b _end\n"
            asm += const_pool
            asm += "_end:\n"
        return asm

    def assemble_recursive(self, expr, function_addr=None, expr_addr=None):
        asm = ""
        const_pool = ""
        ###################################################
        # Complex Expressions
        ###################################################
        ## ldr REG, [CONSTANT]
        if (
            isinstance(expr, ailment.Expr.Load)
            and isinstance(expr.addr, ailment.Expr.Const)
            and expr.addr.value <= 0xFFFF
        ):
            dest = self.reserve_register()
            asm += f"{self.opcode_variant('ldr', size=expr.size)} {dest}, [#{hex(expr.addr.value)}]\n"
            return dest, asm, const_pool
        ## add REG, CONSTANT / lsl REG, CONSTANT
        elif (
            isinstance(expr, ailment.Expr.BinaryOp)
            and isinstance(expr.operands[1], ailment.Expr.Const)
            and expr.operands[1].value <= 0xFFFF
        ):
            operand0, ret_asm, ret_const_pool = self.assemble_recursive(expr.operands[0], function_addr, expr_addr)
            asm += ret_asm
            const_pool += ret_const_pool
            dest = self.reserve_register()
            if expr.op == "Add":
                asm += f"add {dest}, {operand0}, #{hex(expr.operands[1].value)}\n"
            elif expr.op == "Shl":
                asm += f"lsl {dest}, {operand0}, #{hex(expr.operands[1].value)}\n"
            self.return_register(operand0)
            return dest, asm, const_pool
        ## ldr REG, [REG, CONSTANT]
        elif (
            isinstance(expr, ailment.Expr.Load)
            and isinstance(expr.addr, ailment.Expr.BinaryOp)
            and expr.addr.op == "Add"
            and isinstance(expr.addr.operands[1], ailment.Expr.Const)
            and expr.addr.operands[1].value <= 0xFFFF
        ):
            operand0, ret_asm, ret_const_pool = self.assemble_recursive(expr.addr.operands[0], function_addr, expr_addr)
            asm += ret_asm
            const_pool += ret_const_pool
            dest = self.reserve_register()
            asm += (
                f"{self.opcode_variant('ldr', size=expr.size)} {dest}, "
                f"[{operand0}, #{hex(expr.addr.operands[1].value)}]\n"
            )
            self.return_register(operand0)
            return dest, asm, const_pool
        ## ldr REG, [CONSTANT, REG] -> ldr REG, [REG, CONSTANT]
        elif (
            isinstance(expr, ailment.Expr.Load)
            and isinstance(expr.addr, ailment.Expr.BinaryOp)
            and expr.addr.op == "Add"
            and isinstance(expr.addr.operands[0], ailment.Expr.Const)
            and expr.addr.operands[0].value <= 0xFFFF
        ):
            operand1, ret_asm, ret_const_pool = self.assemble_recursive(expr.addr.operands[1], function_addr, expr_addr)
            asm += ret_asm
            const_pool += ret_const_pool
            dest = self.reserve_register()
            asm += (
                f"{self.opcode_variant('ldr', size=expr.size)} {dest}, "
                f"[{operand1}, #{hex(expr.addr.operands[0].value)}]\n"
            )
            self.return_register(operand1)
            return dest, asm, const_pool
        ## ldr REG, [REG, REG]
        elif (
            isinstance(expr, ailment.Expr.Load)
            and isinstance(expr.addr, ailment.Expr.BinaryOp)
            and expr.addr.op == "Add"
        ):
            operand0, ret_asm, ret_const_pool = self.assemble_recursive(expr.addr.operands[0], function_addr, expr_addr)
            asm += ret_asm
            const_pool += ret_const_pool
            operand1, ret_asm, ret_const_pool = self.assemble_recursive(expr.addr.operands[1], function_addr, expr_addr)
            asm += ret_asm
            const_pool += ret_const_pool
            dest = self.reserve_register()
            asm += f"{self.opcode_variant('ldr', size=expr.size)} {dest}, [{operand0}, {operand1}]\n"
            self.return_register(operand0)
            self.return_register(operand1)
            return dest, asm, const_pool
        ## ldr REG, [STACK_OFFSET]
        elif isinstance(expr, ailment.Expr.Load) and isinstance(expr.addr, ailment.Expr.StackBaseOffset):
            dest = self.reserve_register()
            asm += (
                f"{self.opcode_variant('ldr', size=expr.size)} "
                f"{dest}, [sp, "
                f"{self.bp_offset_to_sp_offset(expr.addr.offset, function_addr, expr_addr, self.sp_adjust)}]\n"
            )
            return dest, asm, const_pool

        ###################################################
        # Fall back to the default implementation
        ###################################################
        ## Ignore Convert
        elif isinstance(expr, ailment.Expr.Convert):
            return self.assemble_recursive(expr.operand, function_addr, expr_addr)
        ## ldr REG, [REG]
        elif isinstance(expr, ailment.Expr.Load):
            src, ret_asm, ret_const_pool = self.assemble_recursive(expr.addr, function_addr, expr_addr)
            asm += ret_asm
            const_pool += ret_const_pool
            dest = self.reserve_register()
            asm += f"{self.opcode_variant('ldr', size=expr.size)} {dest}, [{src}]\n"
            self.return_register(src)
            return dest, asm, const_pool
        ## add REG, REG, REG / lsl REG, REG, REG
        elif isinstance(expr, ailment.Expr.BinaryOp):
            operand0, ret_asm, ret_const_pool = self.assemble_recursive(expr.operands[0], function_addr, expr_addr)
            asm += ret_asm
            const_pool += ret_const_pool
            operand1, ret_asm, ret_const_pool = self.assemble_recursive(expr.operands[1], function_addr, expr_addr)
            asm += ret_asm
            const_pool += ret_const_pool
            dest = self.reserve_register()
            if expr.op == "Add":
                asm += f"add {dest}, {operand0}, {operand1}\n"
            elif expr.op == "Shl":
                asm += f"lsl {dest}, {operand0}, {operand1}\n"
            self.return_register(operand0)
            self.return_register(operand1)
            return dest, asm, const_pool
        ## Stack Base Offset (bp+offset) -> add REG, sp, #PLACE_HOLDER
        ## sp + offset
        elif isinstance(expr, ailment.Expr.StackBaseOffset):
            dest = self.reserve_register()
            asm += (
                f"{self.opcode_variant('ldr', size=expr.size)}"
                f"{dest}, [sp, "
                f"{self.bp_offset_to_sp_offset(expr.addr.offset, function_addr, expr_addr, self.sp_adjust)}]\n"
            )
            return dest, asm, const_pool
        ## Constant -> mov REG, #CONSTANT
        elif isinstance(expr, ailment.Expr.Const):
            if expr.value <= 0xFFFF:  # either < 16 bits, or > 16 bits but can be represented in 16 bits
                dest = self.reserve_register()
                asm += f"mov {dest}, #{hex(expr.value)}\n"
                return dest, asm, const_pool
            elif expr.bits == 32:
                dest = self.reserve_register()
                asm += f"ldr {dest}, ptr_{hex(expr.value)}\n"
                const_pool += f"ptr_{hex(expr.value)}:\n"
                const_pool += f"  .long {hex(expr.value)}\n"
                return dest, asm, const_pool
            else:
                raise NotImplementedError(f"Unsupported constant size: {expr.bits}")
        else:
            raise Exception("Unsupported expression")

    def display_output(self, s: str):
        AsmOutput(s, parent=self.workspace.main_window).exec_()



from typing import Optional

from PySide6.QtWidgets import QDialog, QDialogButtonBox, QHBoxLayout, QLabel, QLineEdit, QPushButton, QVBoxLayout


class NameLineEdit(QLineEdit):
    """
    Simple line edit with simple identifier validation.
    """

    def __init__(self, textchanged_callback, parent=None):
        super().__init__(parent)

        self.textChanged.connect(textchanged_callback)

    @property
    def name(self):
        text = self.text()
        if self._is_valid_node_name(text):
            return text.strip()
        return None

    @staticmethod
    def _is_valid_node_name(name):
        return name and " " not in name.strip()


class RenameDialog(QDialog):
    """
    A generic dialog box for renaming something.

    If the user enters a valid name and clicks 'OK', the `result` property will
    contain the resulting string. If a user clicks 'Cancel', the `result`
    property will remain `None`.
    """

    def __init__(self, window_title: str = "Rename", initial_text: str = "", parent=None):
        super().__init__(parent)
        self._initial_text: str = initial_text
        self._name_box: NameLineEdit = None
        self._status_label: QLabel = None
        self._ok_button: QPushButton = None
        self.main_layout: QVBoxLayout = QVBoxLayout()
        self.result: Optional[str] = None
        self._init_widgets()
        self.setLayout(self.main_layout)
        self.setWindowTitle(window_title)

    #
    # Private methods
    #

    def _init_widgets(self):
        name_label = QLabel(self)
        name_label.setText("New name")
        name_box = NameLineEdit(self._on_name_changed, self)
        name_box.setText(self._initial_text)
        name_box.selectAll()
        self._name_box = name_box

        label_layout = QHBoxLayout()
        label_layout.addWidget(name_label)
        label_layout.addWidget(name_box)
        self.main_layout.addLayout(label_layout)

        status_label = QLabel(self)
        self.main_layout.addWidget(status_label)
        self._status_label = status_label

        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self._on_ok_clicked)
        buttons.rejected.connect(self.close)
        self._ok_button = buttons.button(QDialogButtonBox.Ok)
        self._ok_button.setEnabled(False)
        self.main_layout.addWidget(buttons)

    #
    # Event handlers
    #

    def _on_name_changed(self, new_text):  # pylint:disable=unused-argument
        if self._name_box is None:
            # initialization is not done yet
            return

        if self._name_box.name is None:
            # the variable name is invalid
            self._status_label.setText("Invalid")
            self._status_label.setProperty("class", "status_invalid")
            self._ok_button.setEnabled(False)
        else:
            self._status_label.setText("Valid")
            self._status_label.setProperty("class", "status_valid")
            self._ok_button.setEnabled(True)

        self._status_label.style().unpolish(self._status_label)
        self._status_label.style().polish(self._status_label)

    def _on_ok_clicked(self):
        node_name = self._name_box.name
        if node_name is not None:
            self.result = node_name
            self.close()



import logging
from typing import TYPE_CHECKING, Generator, List, Optional, Set, Tuple

from angr import KnowledgeBase
from angr.analyses.reaching_definitions.call_trace import CallTrace
from angr.analyses.reaching_definitions.dep_graph import DepGraph
from angr.analyses.reaching_definitions.external_codeloc import ExternalCodeLocation
from angr.calling_conventions import DEFAULT_CC, SimCC, SimRegArg
from angr.knowledge_plugins import Function
from angr.knowledge_plugins.key_definitions.atoms import Register
from angr.knowledge_plugins.key_definitions.constants import OP_BEFORE
from angr.sim_type import SimType
from PySide6.QtWidgets import QMessageBox

from angrmanagement.logic import GlobalInfo
from angrmanagement.logic.threads import gui_thread_schedule_async

from .job import Job

try:
    import argument_resolver
    from argument_resolver.call_trace_visitor import CallTraceSubject
    from argument_resolver.handlers import StdioHandlers, StdlibHandlers, StringHandlers, handler_factory
    from argument_resolver.transitive_closure import transitive_closures_from_defs
except ImportError:
    argument_resolver = None

if TYPE_CHECKING:
    from angr.analyses.reaching_definitions import ReachingDefinitionsAnalysis
    from angr.knowledge_plugins.key_definitions.atoms import Atom

    from angrmanagement.data.instance import Instance


log = logging.getLogger(name=__name__)


class DependencyAnalysisJob(Job):
    """
    Implements a job for dependency analysis.
    """

    def __init__(self, func_addr=None, func_arg_idx=None):
        super().__init__("DependencyAnalysis")

        self.func_addr: Optional[int] = func_addr
        self.func_arg_idx: Optional[int] = func_arg_idx

    def _get_sink_and_atom(self, inst: "Instance"):
        if self.func_addr is not None:
            sinks = [func for func in inst.kb.functions.values() if func.addr == self.func_addr]
            if not sinks:
                return None, None
            sink: Function = sinks[0]

            if sink.calling_convention is not None:
                cc: SimCC = sink.calling_convention
            else:
                cc: SimCC = DEFAULT_CC[inst.project.arch.name](inst.project.arch)

            # TODO: Handle stack-passing arguments
            # TODO: Handle variadic arguments
            arg_locs = cc.arg_locs(sink.prototype)
            arg = arg_locs[self.func_arg_idx]

            # convert arg into atom
            if isinstance(arg, SimRegArg):
                atom = Register(inst.project.arch.registers[arg.reg_name][0], arg.size)
                return sink, atom
            else:
                raise NotImplementedError

        return None, None

    def _run(self, inst: "Instance"):
        self._progress_callback(0.0)
        self._perform(inst)
        self._progress_callback(100.0)

    def _perform(self, inst: "Instance"):
        if not argument_resolver:
            gui_thread_schedule_async(self._display_import_error)
            return

        self._progress_callback(10.0, text="Extracting sink and atom")
        sink, atom = self._get_sink_and_atom(inst)
        if sink is None:
            # invalid sink setup
            return None

        closures = {}
        excluded_functions: Set[int] = set()
        min_depth = 1
        max_depth = 8
        progress_chunk = 70.0 / (max_depth - min_depth)

        for depth in range(min_depth, max_depth):
            base_progress: float = 30.0 + (depth - min_depth) * progress_chunk
            self._progress_callback(base_progress, text="Calculating reaching definitions... depth %d." % depth)
            # generate RDA observation points
            observation_points = set()
            for pred in inst.cfg.am_obj.get_predecessors(inst.cfg.am_obj.get_any_node(self.func_addr)):
                if pred.instruction_addrs:
                    call_inst_addr = pred.instruction_addrs[-1]
                    observation_point = ("insn", call_inst_addr, OP_BEFORE)
                    observation_points.add(observation_point)

            for idx, total, dep in self._dependencies(
                sink, [(atom, SimType())], inst.project.kb, inst.project, depth, excluded_functions, observation_points
            ):
                self._progress_callback(
                    base_progress + idx / total * progress_chunk,
                    text="Computing transitive closures: %d/%d - depth %d" % (idx + 1, total, depth),
                )

                all_defs = set()
                # find the instructions that call this function
                for pred in inst.cfg.am_obj.get_predecessors(inst.cfg.am_obj.get_any_node(self.func_addr)):
                    if pred.instruction_addrs:
                        call_inst_addr = pred.instruction_addrs[-1]
                        loc = ("insn", call_inst_addr, OP_BEFORE)
                        if loc in dep.observed_results:
                            observed_result = dep.observed_results[loc]
                            defs_ = observed_result.get_definitions_from_atoms([atom])
                            all_defs |= defs_

                try:
                    cc = transitive_closures_from_defs(all_defs, dep.dep_graph)
                except Exception:  # pylint:disable=broad-except
                    log.warning("Exception occurred when computing transitive clousure. Skip.")
                    continue

                # determine if there is any values are marked as coming from Externalog. these values are not resolved
                # within the current call-depth range
                has_external = False
                for _def, graph in closures.items():
                    for node in graph.nodes():
                        if isinstance(node.codeloc, ExternalCodeLocation):
                            # yes!
                            has_external = True
                            break
                    if has_external:
                        break
                if not has_external:
                    # fully resolved - we should exclude this function for future exploration
                    current_function_address = dep.subject.content.current_function_address()
                    log.info(
                        "Exclude function %#x from future slices since the data dependencies are fully resolved.",
                        current_function_address,
                    )
                    excluded_functions.add(current_function_address)

                closures.update(cc)

        # display in the dependencies view
        gui_thread_schedule_async(
            self._display_closures,
            (
                inst,
                atom,
                sink.addr,
                closures,
            ),
        )

        return

    @staticmethod
    def _dependencies(
        subject,
        sink_atoms: List[Tuple["Atom", SimType]],
        kb,
        project,
        max_depth: int,
        excluded_funtions: Set[int],
        observation_points: Set[Tuple],
    ) -> Generator[Tuple[int, int, "ReachingDefinitionsAnalysis"], None, None]:
        Handler = handler_factory(
            [
                StdioHandlers,
                StdlibHandlers,
                StringHandlers,
            ]
        )

        if isinstance(subject, Function):
            sink = subject
        else:
            raise TypeError("Unsupported type of subject %s." % type(subject))

        # peek into the callgraph and discover all functions reaching the sink within N layers of calls, which is
        # determined by the depth parameter
        queue: List[Tuple[CallTrace, int]] = [(CallTrace(sink.addr), 0)]
        starts: Set[CallTrace] = set()
        encountered: Set[int] = set(excluded_funtions)
        while queue:
            trace, curr_depth = queue.pop(0)
            if trace.current_function_address() in starts:
                continue
            caller_func_addr = trace.current_function_address()
            callers: Set[int] = set(kb.functions.callgraph.predecessors(caller_func_addr))
            # remove the functions that we already came across - essentially bypassing recursive function calls
            callers = {addr for addr in callers if addr not in encountered}
            caller_depth = curr_depth + 1
            if caller_depth >= max_depth:
                # reached the depth limit. add them to potential analysis starts
                starts |= {trace.step_back(caller_addr, None, caller_func_addr) for caller_addr in callers}
            else:
                # add them to the queue
                for item in (
                    (trace.step_back(caller_addr, None, caller_func_addr), caller_depth) for caller_addr in callers
                ):
                    queue.append(item)
            encountered |= callers

        log.info("Discovered %d function starts at call-depth %d for sink %r.", len(starts), max_depth, sink)

        for idx, start in enumerate(starts):
            handler = Handler(project, sink_function=sink, sink_atoms=sink_atoms, cfg=kb.cfgs[0])
            try:
                rda = project.analyses.ReachingDefinitions(
                    subject=CallTraceSubject(start, kb.functions[start.current_function_address()]),
                    observe_all=False,
                    observation_points=observation_points,
                    function_handler=handler,
                    kb=kb,
                    dep_graph=DepGraph(),
                )
            except Exception:  # pylint:disable=broad-except
                log.warning("Failed to compute dependencies for function %s.", start, exc_info=True)
                continue
            yield idx, len(starts), rda

    @staticmethod
    def _get_new_kb_with_cfgs_and_functions(project, kb):
        new_kb = KnowledgeBase(project)

        new_kb.cfgs = kb.cfgs.copy()
        new_kb.functions = kb.functions.copy()
        new_kb.labels = kb.labels.copy()
        new_kb.resolved_indirect_jumps = kb.resolved_indirect_jumps.copy()
        new_kb.unresolved_indirect_jumps = kb.unresolved_indirect_jumps.copy()

        return new_kb

    @staticmethod
    def _display_import_error():
        QMessageBox.critical(
            None,
            "Import error",
            "Failed to import argument_resolver package. Is operation-mango installed?",
        )

    @staticmethod
    def _display_closures(inst, sink_atom: "Atom", sink_addr: int, closures):
        view = GlobalInfo.main_window.workspace.view_manager.first_view_in_category("dependencies")
        if view is None:
            return

        view.sink_atom = sink_atom
        view.sink_ins_addr = sink_addr
        view.closures = closures
        try:
            view.reload()
        except Exception:
            log.warning("An error occurred when displaying the closures.", exc_info=True)
        GlobalInfo.main_window.workspace.view_manager.raise_view(view)



# Copied and adapted from tendo

import logging
import os
import sys
import tempfile
import unittest
from multiprocessing import Process

from angr.utils.mp import Initializer

logger = logging.getLogger(name=__name__)


if sys.platform != "win32":
    import fcntl


class SingleInstanceException(BaseException):
    pass


class SingleInstance:

    """Class that can be instantiated only once per machine.

    If you want to prevent your script from running in parallel just instantiate SingleInstance() class. If is there
    another instance already running it will throw a `SingleInstanceException`.

    >>> me = SingleInstance()

    This option is very useful if you have scripts executed by crontab at small amounts of time.

    Remember that this works by creating a lock file with a filename based on the full path to the script file.

    Providing a flavor_id will augment the filename with the provided flavor_id, allowing you to create multiple
    singleton instances from the same file. This is particularly useful if you want specific functions to have their
    own singleton instances.
    """

    def __init__(self, flavor_id="", lockfile=""):
        self.initialized = False
        if lockfile:
            self.lockfile = lockfile
        else:
            basename = (
                os.path.splitext(os.path.abspath(__file__))[0].replace("/", "-").replace(":", "").replace("\\", "-")
                + "-%s" % flavor_id
                + ".lock"
            )
            self.lockfile = os.path.normpath(tempfile.gettempdir() + "/" + basename)

        logger.debug("SingleInstance lockfile: %s", self.lockfile)
        if sys.platform == "win32":
            try:
                # file already exists, we try to remove (in case previous
                # execution was interrupted)
                if os.path.exists(self.lockfile):
                    os.unlink(self.lockfile)
                self.fd = os.open(self.lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)
            except OSError as ex:
                type, e, tb = sys.exc_info()
                if e.errno == 13:
                    logger.debug("Another instance is already running, quitting.")
                    raise SingleInstanceException from ex
                print(e.errno)
                raise
        else:  # non Windows
            self.fp = open(self.lockfile, "w")
            self.fp.flush()
            try:
                fcntl.lockf(self.fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except OSError as ex:
                logger.debug("Another instance is already running, quitting.")
                raise SingleInstanceException from ex
        self.initialized = True

    def __del__(self):
        if not self.initialized:
            return
        try:
            if sys.platform == "win32":
                if hasattr(self, "fd"):
                    os.close(self.fd)
                    os.unlink(self.lockfile)
            else:
                fcntl.lockf(self.fp, fcntl.LOCK_UN)
                # os.close(self.fp)
                if os.path.isfile(self.lockfile):
                    os.unlink(self.lockfile)
        except Exception as e:
            if logger:
                logger.debug(e)
            else:
                print("Unloggable error: %s" % e)
            sys.exit(-1)


def f(initializer: Initializer, name):
    initializer.initialize()
    tmp = logger.level
    logger.setLevel(logging.CRITICAL)  # we do not want to see the warning
    try:
        me2 = SingleInstance(flavor_id=name)  # noqa
    except SingleInstanceException:
        sys.exit(-1)
    logger.setLevel(tmp)
    pass


class testSingleton(unittest.TestCase):
    def test_1(self):
        me = SingleInstance(flavor_id="test-1")
        del me  # now the lock should be removed
        assert True

    def test_2(self):
        p = Process(
            target=f,
            args=(
                Initializer.get(),
                "test-2",
            ),
        )
        p.start()
        p.join()
        # the called function should succeed
        assert p.exitcode == 0, "%s != 0" % p.exitcode

    def test_3(self):
        me = SingleInstance(flavor_id="test-3")  # noqa -- me should still kept
        p = Process(
            target=f,
            args=(
                Initializer.get(),
                "test-3",
            ),
        )
        p.start()
        p.join()
        # the called function should fail because we already have another
        # instance running
        assert p.exitcode != 0, "%s != 0 (2nd execution)" % p.exitcode
        # note, we return -1 but this translates to 255 meanwhile we'll
        # consider that anything different from 0 is good
        p = Process(
            target=f,
            args=(
                Initializer.get(),
                "test-3",
            ),
        )
        p.start()
        p.join()
        # the called function should fail because we already have another
        # instance running
        assert p.exitcode != 0, "%s != 0 (3rd execution)" % p.exitcode

    def test_4(self):
        lockfile = "/tmp/foo.lock"
        me = SingleInstance(lockfile=lockfile)
        assert me.lockfile == lockfile


if __name__ == "__main__":
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.DEBUG)
    unittest.main()



from .debug_toolbar import DebugToolbar
from .feature_map_toolbar import FeatureMapToolbar
from .file_toolbar import FileToolbar
from .function_table_toolbar import FunctionTableToolbar
from .nav_toolbar import NavToolbar

__all__ = [
    "DebugToolbar",
    "FeatureMapToolbar",
    "FileToolbar",
    "FunctionTableToolbar",
    "NavToolbar",
]



import logging
import os
import time
import traceback
from typing import TYPE_CHECKING, Callable, List, Optional, Type, TypeVar, Union

from angr import StateHierarchy
from angr.knowledge_plugins.cfg import MemoryData, MemoryDataSort
from angr.knowledge_plugins.functions.function import Function
from angr.knowledge_plugins.patches import Patch
from angr.misc.testing import is_testing
from cle import SymbolType
from PySide6.QtWidgets import QMessageBox

from angrmanagement.config import Conf
from angrmanagement.data.analysis_options import (
    AnalysesConfiguration,
    CFGAnalysisConfiguration,
    CodeTaggingConfiguration,
    FlirtAnalysisConfiguration,
    VariableRecoveryConfiguration,
)
from angrmanagement.data.breakpoint import Breakpoint, BreakpointType
from angrmanagement.data.instance import ObjectContainer
from angrmanagement.data.jobs import (
    CFGGenerationJob,
    CodeTaggingJob,
    FlirtSignatureRecognitionJob,
    Job,
    PrototypeFindingJob,
    VariableRecoveryJob,
)
from angrmanagement.data.jobs.loading import LoadBinaryJob
from angrmanagement.data.trace import BintraceTrace, Trace
from angrmanagement.logic.commands import CommandManager
from angrmanagement.logic.debugger import DebuggerWatcher
from angrmanagement.logic.debugger.bintrace import BintraceDebugger
from angrmanagement.logic.debugger.simgr import SimulationDebugger
from angrmanagement.logic.threads import gui_thread_schedule_async
from angrmanagement.plugins import PluginManager
from angrmanagement.ui.dialogs import AnalysisOptionsDialog
from angrmanagement.utils import locate_function
from angrmanagement.utils.daemon_thread import start_daemon_thread

from .view_manager import ViewManager
from .views import (
    BaseView,
    BreakpointsView,
    CallExplorerView,
    CodeView,
    ConsoleView,
    DataDepView,
    DependencyView,
    DisassemblyView,
    FunctionsView,
    HexView,
    InteractionView,
    LogView,
    PatchesView,
    ProximityView,
    RegistersView,
    StackView,
    StatesView,
    StringsView,
    SymexecView,
    TraceMapView,
    TracesView,
    TypesView,
)

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.main_window import MainWindow


_l = logging.getLogger(__name__)
T = TypeVar("T")


class Workspace:
    """
    This class implements the angr management workspace.
    """

    def __init__(self, main_window, instance):
        self.main_window: MainWindow = main_window
        self._main_instance = instance
        instance.workspace = self

        self.command_manager: CommandManager = CommandManager()
        self.view_manager: ViewManager = ViewManager(self)
        self.plugins: PluginManager = PluginManager(self)
        self.variable_recovery_job: Optional[VariableRecoveryJob] = None
        self._first_cfg_generation_callback_completed: bool = False

        # Configure callbacks on main_instance
        instance.project.am_subscribe(self._instance_project_initalization)
        instance.simgrs.am_subscribe(self._update_simgr_debuggers)
        instance.handle_comment_changed_callback = self.plugins.handle_comment_changed
        instance.job_worker_exception_callback = self._handle_job_exception

        self.current_screen = ObjectContainer(None, name="current_screen")

        self.default_tabs = [
            DisassemblyView(self, self._main_instance, "center"),
            HexView(self, self._main_instance, "center"),
            CodeView(self, self._main_instance, "center"),
            FunctionsView(self, self._main_instance, "left"),
        ]
        if Conf.has_operation_mango:
            self.default_tabs.append(DependencyView(self, self._main_instance, "center"))
        self.default_tabs += [
            ConsoleView(self, self._main_instance, "bottom"),
            LogView(self, self._main_instance, "bottom"),
        ]

        enabled_tabs = [x.strip() for x in Conf.enabled_tabs.split(",") if x.strip()]
        for tab in self.default_tabs:
            if tab.__class__.__name__ in enabled_tabs or len(enabled_tabs) == 0:
                self.add_view(tab)

        self._dbg_watcher = DebuggerWatcher(self.on_debugger_state_updated, self.main_instance.debugger_mgr.debugger)
        self.on_debugger_state_updated()

        DisassemblyView.register_commands(self)

        instance.patches.am_subscribe(self._on_patch_event)

    #
    # Properties
    #

    @property
    def _main_window(self) -> "MainWindow":
        return self.main_window

    @property
    def main_instance(self) -> "Instance":
        return self._main_instance

    #
    # Events
    #

    def on_debugger_state_updated(self):
        """
        Jump to debugger target PC in active disassembly view.
        """
        # FIXME: the disassembly view should subscribe to debugger updates, but for that we will need to expose
        #        a mechanism for the view to select between states. For now we simply have a global debugger
        #        selection.
        dbg = self._dbg_watcher.debugger
        if not dbg.am_none:
            state = dbg.simstate
            if state is not None:
                addr = state.solver.eval(state.regs.pc)
                view = self.view_manager.current_view_in_category(
                    "disassembly"
                ) or self.view_manager.first_view_in_category("disassembly")
                if view:
                    view.jump_to(addr, True)

    def on_function_selected(self, func: Function):
        """
        Callback function triggered when a new function is selected in the function view.

        :param func:    The function that is selected.
        :return:        None
        """

        # Ask all current views to display this function
        current_view = self.view_manager.current_tab
        if current_view is None or not current_view.FUNCTION_SPECIFIC_VIEW:
            # we don't have a current view or the current view does not have function-specific content. create a
            # disassembly view to display the selected function.
            disasm_view = self._get_or_create_view("dissasembly", DisassemblyView)
            disasm_view.display_function(func)
            self.view_manager.raise_view(disasm_view)
        else:
            # ask the current view to display this function
            current_view.function = func

    def on_function_tagged(self):
        # reload disassembly view
        if len(self.view_manager.views_by_category["disassembly"]) == 1:
            view = self.view_manager.first_view_in_category("disassembly")
        else:
            view = self.view_manager.current_view_in_category("disassembly")

        if view is not None and view.current_function.am_obj is not None:
            view.reload()

    def on_variable_recovered(self, func_addr: int):
        """
        Called when variable information of the given function is available.

        :param int func_addr:   Address of the function whose variable information is available.
        """
        disassembly_view = self.view_manager.first_view_in_category("disassembly")
        if disassembly_view is not None:
            disassembly_view.on_variable_recovered(func_addr)

    def generate_cfg(self, cfg_args=None):
        if cfg_args is None:
            cfg_args = {}

        cfg_job = CFGGenerationJob(on_finish=self.on_cfg_generated, **cfg_args)
        self.main_instance.add_job(cfg_job)
        start_daemon_thread(self._refresh_cfg, "Progressively Refreshing CFG", args=(cfg_job,))

    def _refresh_cfg(self, cfg_job):
        """
        Reload once and then refresh in a loop, while the CFG job is running
        """
        reloaded = False
        while True:
            if not self.main_instance.cfg.am_none:
                if reloaded:
                    gui_thread_schedule_async(
                        self.refresh,
                        kwargs={
                            "categories": ["disassembly", "functions"],
                        },
                    )
                else:
                    gui_thread_schedule_async(
                        self.reload,
                        kwargs={
                            "categories": ["disassembly", "functions"],
                        },
                    )
                    reloaded = True

            time.sleep(0.3)
            if cfg_job not in self.main_instance.jobs:
                break

    def on_cfg_generated(self):
        if self.main_instance._analysis_configuration["flirt"].enabled:
            self.main_instance.add_job(
                FlirtSignatureRecognitionJob(
                    on_finish=self._on_flirt_signature_recognized,
                )
            )

        if self.main_instance.cfg is not None:
            if not self._first_cfg_generation_callback_completed:
                self._first_cfg_generation_callback_completed = True
                the_func = self.main_instance.kb.functions.function(name="main")
                if the_func is None:
                    the_func = self.main_instance.kb.functions.function(addr=self.main_instance.project.entry)
                if the_func is not None:
                    self.on_function_selected(the_func)

            # Reload the pseudocode view
            view = self.view_manager.first_view_in_category("pseudocode")
            if view is not None:
                view.reload()

            # Reload the strings view
            view = self.view_manager.first_view_in_category("strings")
            if view is not None:
                view.reload()

            # Clear the proximity view
            view = self.view_manager.first_view_in_category("proximity")
            if view is not None:
                view.clear()

    def _on_flirt_signature_recognized(self):
        self.main_instance.add_job(
            PrototypeFindingJob(
                on_finish=self._on_prototype_found,
            )
        )

    def _on_prototype_found(self):
        if self.main_instance._analysis_configuration["code_tagging"].enabled:
            self.main_instance.add_job(
                CodeTaggingJob(
                    on_finish=self.on_function_tagged,
                )
            )

        if self.main_instance._analysis_configuration["varec"].enabled:
            options = self.main_instance._analysis_configuration["varec"].to_dict()
            if is_testing:
                # disable multiprocessing on angr CI
                options["workers"] = 0
            self.main_instance.variable_recovery_job = VariableRecoveryJob(
                **self.main_instance._analysis_configuration["varec"].to_dict(),
                on_variable_recovered=self.on_variable_recovered,
            )
            # prioritize the current function in display
            disassembly_view = self.view_manager.first_view_in_category("disassembly")
            if disassembly_view is not None and not disassembly_view.function.am_none:
                self.main_instance.variable_recovery_job.prioritize_function(disassembly_view.function.addr)
            self.main_instance.add_job(self.main_instance.variable_recovery_job)

    def _on_patch_event(self, **kwargs):
        if self.main_instance.cfg.am_none:
            return

        update_cfg = False
        for k in ("added", "removed"):
            for patch in kwargs.get(k, set()):
                self.main_instance.cfg.clear_region_for_reflow(
                    patch.addr, len(patch.new_bytes), self.main_instance.project.kb
                )
                update_cfg = True

        if update_cfg:
            self.generate_cfg(
                cfg_args={
                    "force_smart_scan": False,
                    "force_complete_scan": False,
                    "model": self.main_instance.kb.cfgs.get_most_accurate(),
                    # FIXME: We don't want to force scan the entire binary, just the patched region. Add an
                    #        option for it.
                }
            )

    #
    # Public methods
    #

    def new_disassembly_view(self) -> DisassemblyView:
        """
        Add a new disassembly view.
        """
        disassembly_view = self.view_manager.first_view_in_category("disassembly")
        current_addr = disassembly_view.jump_history.current if disassembly_view is not None else None

        view = DisassemblyView(self, self._main_instance, "center")
        self.add_view(view)
        self.raise_view(view)
        view._linear_viewer.initialize()  # FIXME: Don't access protected member
        if current_addr is not None:
            view.jump_to(current_addr)
        # TODO move view tab to front of dock
        return view

    def add_view(self, view):
        self.view_manager.add_view(view)

    def remove_view(self, view):
        self.view_manager.remove_view(view)

    def raise_view(self, view):
        """
        Find the dock widget of a view, and then bring that dockable to front.

        :param BaseView view: The view to raise.
        :return:              None
        """

        self.view_manager.raise_view(view)

    def reload(self, categories: Optional[List[str]] = None):
        """
        Ask all or specified views to reload the underlying data and regenerate the UI. This is usually expensive.

        :param categories:  Specify a list of view categories that should be reloaded.
        :return:            None
        """

        if categories is None:
            views = self.view_manager.views
        else:
            views = []
            for category in categories:
                views.extend(self.view_manager.views_by_category.get(category, []))

        for view in views:
            try:
                view.reload()
            except Exception:  # pylint:disable=broad-except
                _l.warning("Exception occurred during reloading view %s.", view, exc_info=True)

    def refresh(self, categories: Optional[List[str]] = None):
        """
        Ask all or specified views to refresh based on changes in the underlying data and refresh the UI if needed. This
        may be called frequently so it must be extremely fast.

        :param categories:  Specify a list of view categories that should be reloaded.
        :return:            None
        """

        if categories is None:
            views = self.view_manager.views
        else:
            views = []
            for category in categories:
                views.extend(self.view_manager.views_by_category.get(category, []))

        for view in views:
            try:
                view.refresh()
            except Exception:  # pylint:disable=broad-except
                _l.warning("Exception occurred during reloading view %s.", view, exc_info=True)

    def viz(self, obj):
        """
        Visualize the given object.

        - For integers, open the disassembly view and jump to that address
        - For Function objects, open the disassembly view and jump there
        - For strings, look up the symbol of that name and jump there
        """

        if isinstance(obj, int):
            self.jump_to(obj)
        elif isinstance(obj, str):
            sym = self.main_instance.project.loader.find_symbol(obj)
            if sym is not None:
                self.jump_to(sym.rebased_addr)
        elif isinstance(obj, Function):
            self.jump_to(obj.addr)

    def jump_to(self, addr, view=None, use_animation=False):
        if view is None or view.category != "disassembly":
            view = self._get_or_create_view("disassembly", DisassemblyView)

        self.raise_view(view)
        view.setFocus()
        view.jump_to(addr, use_animation=use_animation)

    def add_breakpoint(self, obj: Union[str, int], type_: Optional[str] = None, size: Optional[int] = None):
        """
        Convenience function to add a breakpoint.

        Examples:
        - `workspace.add_breakpoint(0x1234)` sets an execution breakpoint on address 0x1234
        - `workspace.add_breakpoint('main')` sets an execution breakpoint on `main` function
        - `workspace.add_breakpoint('global_value')` sets a write breakpoint on `global_value`
        - `workspace.add_breakpoint('global_value', 'read', 1)` sets a 1-byte read breakpoint on `global_value`
        """
        if isinstance(obj, int):
            addr = obj
        elif isinstance(obj, str):
            sym = self.main_instance.project.loader.find_symbol(obj)
            if sym is None:
                _l.error("Couldn't resolve '%s'", obj)
                return
            addr = sym.rebased_addr
            if not size:
                size = sym.size
            if not type_:
                type_ = "execute" if sym.type == SymbolType.TYPE_FUNCTION else "write"
        elif isinstance(obj, Function):
            addr = obj.addr
            if not type_:
                type_ = "execute"
        else:
            _l.error("Unexpected target object type. Expected int | str | Function")
            return

        if not size:
            size = 1

        bp_type_map = {
            None: BreakpointType.Execute,
            "execute": BreakpointType.Execute,
            "write": BreakpointType.Write,
            "read": BreakpointType.Read,
        }
        if type_ not in bp_type_map:
            _l.error("Unknown breakpoint type '%s'. Expected %s", type_, " | ".join(bp_type_map.keys()))
            return

        bp = Breakpoint(bp_type_map[type_], addr, size)
        self.main_instance.breakpoint_mgr.add_breakpoint(bp)

    def set_comment(self, addr, comment_text):
        self.main_instance.set_comment(addr, comment_text)

        disasm_view = self._get_or_create_view("disassembly", DisassemblyView)
        if disasm_view._flow_graph.disasm is not None:
            # redraw
            disasm_view.current_graph.refresh()

    def run_analysis(self, prompt_for_configuration=True):
        if self.main_instance.project.am_none:
            return

        if not self.main_window.shown_at_start:
            # If we are running headlessly (e.g. tests), just run with default configuration
            prompt_for_configuration = False

        if self.main_instance._analysis_configuration is None:
            self.main_instance._analysis_configuration = AnalysesConfiguration(
                [
                    a(self.main_instance)
                    for a in [
                        CFGAnalysisConfiguration,
                        FlirtAnalysisConfiguration,
                        CodeTaggingConfiguration,
                        VariableRecoveryConfiguration,
                    ]
                ],
                self,
            )

        if prompt_for_configuration:
            dlg = AnalysisOptionsDialog(self.main_instance._analysis_configuration, self, self.main_window)
            dlg.setModal(True)
            should_run = dlg.exec_()
        else:
            should_run = True

        if should_run and self.main_instance._analysis_configuration["cfg"].enabled:
            cfg_options = self.main_instance._analysis_configuration["cfg"].to_dict()
            # update function start locations
            if "function_starts" in cfg_options:
                function_starts = []
                for func_start_str in cfg_options["function_starts"].split(","):
                    func_start_str = func_start_str.strip(" ")
                    if not func_start_str:
                        continue

                    try:
                        func_addr = int(func_start_str, 16)
                    except ValueError:
                        if prompt_for_configuration:
                            QMessageBox.critical(
                                None, "Invalid function start string", f"Invalid analysis start {func_start_str}."
                            )
                        return

                    function_starts.append(func_addr)

                if function_starts:
                    if "explicit_analysis_starts" in cfg_options:
                        cfg_options["elf_eh_frame"] = False
                        cfg_options["symbols"] = False
                        cfg_options["start_at_entry"] = False

                    cfg_options["function_starts"] = function_starts

            # discard "explicit_analysis_starts" even if function_starts is not set
            if "explicit_analysis_starts" in cfg_options:
                del cfg_options["explicit_analysis_starts"]

            # update options for region specification
            if "regions" in cfg_options:
                regions = []
                for region_str in cfg_options["regions"].split(","):
                    region_str = region_str.strip(" ")
                    if not region_str:
                        continue
                    if "-" not in region_str or region_str.count("-") != 1:
                        # invalid region
                        if prompt_for_configuration:
                            QMessageBox.critical(
                                None, "Invalid region setting", f"Invalid analysis region {region_str}."
                            )
                        return
                    min_addr, max_addr = region_str.split("-")
                    try:
                        min_addr = int(min_addr, 16)
                    except ValueError:
                        if prompt_for_configuration:
                            QMessageBox.critical(
                                None, "Invalid region setting", f"Invalid analysis region {region_str}."
                            )
                        return
                    try:
                        max_addr = int(max_addr, 16)
                    except ValueError:
                        if prompt_for_configuration:
                            QMessageBox.critical(
                                None, "Invalid region setting", f"Invalid analysis region {region_str}."
                            )
                        return
                    regions.append((min_addr, max_addr))
                if regions:
                    cfg_options["regions"] = regions

            self.generate_cfg(cfg_options)

    def decompile_current_function(self):
        current = self.view_manager.current_tab
        if isinstance(current, CodeView):
            current.decompile()
        else:
            view = self._get_or_create_view("disassembly", DisassemblyView)
            view.decompile_current_function()

    def view_data_dependency_graph(self, analysis_params: dict):
        view = self._get_or_create_view("data_dependency", DataDepView)
        view.analysis_params = analysis_params
        self.raise_view(view)

    def view_proximity_for_current_function(self, view=None):
        if view is None or view.category != "proximity":
            view = self._get_or_create_view("proximity", ProximityView)

        disasm_view = self._get_or_create_view("disassembly", DisassemblyView)
        if disasm_view.current_function is not None:
            view.function = disasm_view.current_function.am_obj

        self.raise_view(view)

    def decompile_function(self, func: Function, curr_ins=None, view=None):
        """
        Decompile a function and switch to decompiled view. If curr_ins is
        defined, then also switch cursor focus to the position associated
        with the asm instruction addr

        :param func: The function to decompile
        :param curr_ins: The instruction the cursor was at before switching to decompiled view
        :param view: The decompiled qt text view
        :return:
        """

        if view is None or view.category != "pseudocode":
            view = self._get_or_create_view("pseudocode", CodeView)

        view.function.am_obj = func
        view.function.am_event(focus=True, focus_addr=curr_ins)

    def create_simulation_manager(self, state, state_name, view=None):
        inst = self.main_instance
        hierarchy = StateHierarchy()
        simgr = inst.project.factory.simulation_manager(state, hierarchy=hierarchy)
        simgr_container = ObjectContainer(simgr, name=state_name)
        inst.simgrs.append(simgr_container)
        inst.simgrs.am_event(src="new_path")

        if view is None:
            view = self._get_or_create_view("symexec", SymexecView)
        view.select_simgr(simgr_container)

        self.raise_view(view)

    def create_trace_debugger(self):
        if self.main_instance.current_trace.am_none:
            _l.error("No trace available")
            return

        if isinstance(self.main_instance.current_trace.am_obj, BintraceTrace):
            dbg = BintraceDebugger(self.main_instance.current_trace.am_obj, self)
            dbg.init()
            self.main_instance.debugger_list_mgr.add_debugger(dbg)
            self.main_instance.debugger_mgr.set_debugger(dbg)

    def is_current_trace(self, trace: Optional[Trace]) -> bool:
        return self.main_instance.current_trace.am_obj is trace

    def set_current_trace(self, trace: Optional[Trace]):
        self.main_instance.current_trace.am_obj = trace
        self.main_instance.current_trace.am_event()

    def remove_trace(self, trace: Trace):
        if self.is_current_trace(trace):
            self.set_current_trace(None)
        self.main_instance.traces.remove(trace)
        self.main_instance.traces.am_event(trace_removed=trace)
        # Note: Debuggers and other objects may retain trace

    def load_trace_from_path(self, path: str):
        if not BintraceTrace.trace_backend_enabled():
            QMessageBox.critical(None, "Error", "bintrace is not installed. Please install the bintrace package.")
            return

        _l.info("Opening trace %s", path)
        trace = BintraceTrace.load_trace(path)

        def on_complete():
            if not self.main_instance.project.am_none:
                self.main_instance.traces.append(trace)
                self.main_instance.traces.am_event(trace_added=trace)
                self.main_instance.current_trace.am_obj = trace
                self.main_instance.current_trace.am_event()
                self.show_traces_view()

        if self.main_instance.project.am_none:
            QMessageBox.information(None, "Creating project", "Creating new project from trace.")
            self.create_project_from_trace(trace, on_complete)
        else:
            on_complete()

    def create_project_from_trace(self, trace: Trace, on_complete: Callable):
        thing = None
        load_options = trace.get_project_load_options()

        if load_options is None:
            QMessageBox.warning(
                None, "Error", "Failed to determine load options for binary from trace. Please select binary."
            )
            load_options = {}
        else:
            thing = load_options["thing"]
            load_options = load_options["load_options"]
            if not os.path.exists(thing):
                QMessageBox.warning(
                    None,
                    "Unable to find target binary!",
                    f"Unable to find the traced binary at: \n\n{thing}\n\nPlease select target binary.",
                )
                thing = None

        if not thing:
            thing = self.main_window.open_mainfile_dialog()

        if not thing:
            on_complete()
            return

        self.main_instance.binary_path = thing
        self.main_instance.original_binary_path = thing
        job = LoadBinaryJob(thing, load_options=load_options, on_finish=on_complete)
        self.main_instance.add_job(job)

    def interact_program(self, img_name, view=None):
        if view is None or view.category != "interaction":
            view = self._get_or_create_view("interaction", InteractionView)
        view.initialize(img_name)

        self.raise_view(view)
        view.setFocus()

    def log(self, msg):
        if isinstance(msg, BaseException):
            msg = "".join(traceback.format_exception(type(msg), msg, msg.__traceback__))

        console = self.view_manager.first_view_in_category("console")
        if console is None or not console.ipython_widget_available:
            print(msg)
        else:
            console.print_text(msg)
            console.print_text("\n")

    def show_view(self, category: str, type_: Type[BaseView], position: str = "center"):
        view = self._get_or_create_view(category, type_, position=position)
        self.raise_view(view)
        view.setFocus()

    def show_linear_disassembly_view(self):
        view = self._get_or_create_view("disassembly", DisassemblyView, position="center")
        view.display_linear_viewer()
        self.raise_view(view)
        view.setFocus()

    def show_graph_disassembly_view(self):
        view = self._get_or_create_view("disassembly", DisassemblyView, position="center")
        view.display_disasm_graph()
        self.raise_view(view)
        view.setFocus()

    def create_and_show_linear_disassembly_view(self):
        """
        Create a new disassembly view and select the Linear disassembly mode.
        """
        view = self.new_disassembly_view()
        view.display_linear_viewer()
        self.raise_view(view)
        view.setFocus()

    def create_and_show_graph_disassembly_view(self):
        """
        Create a new disassembly view and select the Graph disassembly mode.
        """
        view = self.new_disassembly_view()
        view.display_disasm_graph()
        self.raise_view(view)
        view.setFocus()

    def show_pseudocode_view(self):
        self.show_view("pseudocode", CodeView)

    def show_hex_view(self):
        self.show_view("hex", HexView)

    def show_symexec_view(self):
        self.show_view("symexec", SymexecView)

    def show_states_view(self):
        self.show_view("states", StatesView)

    def show_strings_view(self):
        self.show_view("strings", StringsView)

    def show_patches_view(self):
        self.show_view("patches", PatchesView)

    def show_interaction_view(self):
        self.show_view("interaction", InteractionView)

    def show_types_view(self):
        self.show_view("types", TypesView)

    def show_functions_view(self):
        self.show_view("functions", FunctionsView, position="left")

    def show_traces_view(self):
        self.show_view("traces", TracesView)

    def show_trace_map_view(self):
        self.show_view("tracemap", TraceMapView, position="top")

    def show_registers_view(self):
        self.show_view("registers", RegistersView, position="right")

    def show_stack_view(self):
        self.show_view("stack", StackView, position="right")

    def show_breakpoints_view(self):
        self.show_view("breakpoints", BreakpointsView)

    def show_call_explorer_view(self):
        self.show_view("call_explorer", CallExplorerView)

    def show_console_view(self):
        self.show_view("console", ConsoleView, position="bottom")

    def show_log_view(self):
        self.show_view("log", LogView, position="bottom")

    def create_and_show_hex_view(self):
        view = HexView(self._main_instance, "center")
        self.add_view(view)
        return view

    def toggle_exec_breakpoint(self):
        if self.main_instance is None:
            return

        view: Optional[DisassemblyView] = self.view_manager.first_view_in_category("disassembly")
        if view is not None:
            selected_insns = view.current_graph.infodock.selected_insns
            if selected_insns:
                for insn in selected_insns:
                    self.main_instance.breakpoint_mgr.toggle_exec_breakpoint(insn)

    def step_forward(self, until_addr: Optional[int] = None):
        if self.main_instance is None:
            return

        self.main_instance.debugger_mgr.debugger.step_forward(until_addr=until_addr)

    def continue_forward(self):
        if self.main_instance is None:
            return

        self.main_instance.debugger_mgr.debugger.continue_forward()

    def append_code_to_console(self, hook_code_string):
        console = self._get_or_create_view("console", ConsoleView)
        console.set_input_buffer(hook_code_string)

    def patch(self, addr: int, asm: str, pad: bool = True) -> None:
        ks = self.main_instance.project.arch.keystone
        block = self.main_instance.project.factory.block(addr)
        insn = block.disassembly.insns[0]
        original_bytes: bytes = self.main_instance.project.loader.memory.load(insn.address, insn.size)
        ks = self.main_instance.project.arch.keystone
        new_bytes = (ks.asm(asm, addr, as_bytes=True)[0] or b"") if len(asm) else b""

        # Pad to original instruction length
        byte_length_delta = len(original_bytes) - len(new_bytes)
        if byte_length_delta > 0:
            if pad:
                nop_instruction_bytes = self.main_instance.project.arch.nop_instruction
                new_bytes += (byte_length_delta // len(nop_instruction_bytes)) * nop_instruction_bytes
                byte_length_delta = len(original_bytes) - len(new_bytes)
                if byte_length_delta:
                    _l.warning("Unable to completely pad remainder")
        elif byte_length_delta < 0:
            _l.warning("Patch exceeds original instruction length")

        pm = self.main_instance.project.kb.patches
        patch = Patch(addr, new_bytes)
        pm.add_patch_obj(patch)
        self.main_instance.patches.am_event(added={patch})

    def define_code(self, addr: int):
        cfg = self.main_instance.cfg
        if cfg.am_none:
            _l.error("Run initial CFG analysis before defining code")
            return

        func = locate_function(self.main_instance, addr)
        if func is not None:
            _l.warning("Address %#x is already defined as code", addr)
            return

        # Attempt flow into preceding function
        func = cfg.find_function_for_reflow_into_addr(addr)
        if func:
            cfg.clear_region_for_reflow(func.addr)

        # Truncate existing memory data
        if addr in cfg.memory_data:
            del cfg.memory_data[addr]
        for md in cfg.memory_data.values():
            if md.size and md.addr < addr < (md.addr + md.size):
                md.size = addr - md.addr

        self.generate_cfg(
            cfg_args={
                "symbols": False,
                "function_prologues": False,
                "start_at_entry": False,
                "force_smart_scan": False,
                "force_complete_scan": False,
                "function_starts": [func.addr if func else addr],
                "model": self.main_instance.kb.cfgs.get_most_accurate(),
            }
        )

    def undefine_code(self, addr: int):
        cfg = self.main_instance.cfg
        if cfg.am_none:
            _l.error("Run initial CFG analysis before undefining code")
            return

        func = locate_function(self.main_instance, addr)
        if func is None:
            _l.warning("Could not determine function for addr %#x", addr)
            return

        md = MemoryData(addr, 1, MemoryDataSort.Integer)  # FIXME: Type, expand size
        cfg.memory_data[md.addr] = md.copy()
        cfg.clear_region_for_reflow(func.addr)

        self.generate_cfg(
            cfg_args={
                "symbols": False,
                "function_prologues": False,
                "start_at_entry": False,
                "force_smart_scan": False,
                "force_complete_scan": False,
                "function_starts": [func.addr],
                "model": self.main_instance.kb.cfgs.get_most_accurate(),
            }
        )

    #
    # Instance Callbacks
    #
    def _instance_project_initalization(self, **kwargs):  # pylint:disable=unused-argument
        if self.main_instance.project.am_none:
            return

        gui_thread_schedule_async(self.run_analysis)

        self.plugins.handle_project_initialization()

    def _handle_job_exception(self, job: Job, e: Exception):
        self.log('Exception while running job "%s":' % job.name)
        self.log(e)
        self.log("Type %debug to debug it")

    def _update_simgr_debuggers(self, **kwargs):  # pylint:disable=unused-argument
        sim_dbg = None
        for dbg in self.main_instance.debugger_list_mgr.debugger_list:
            if isinstance(dbg, SimulationDebugger):
                sim_dbg = dbg
                break

        if len(self.main_instance.simgrs) > 0:
            if sim_dbg is None:
                simgrs = self._get_or_create_view("symexec", SymexecView)._simgrs
                dbg = SimulationDebugger(simgrs, self)
                self.main_instance.debugger_list_mgr.add_debugger(dbg)
                self.main_instance.debugger_mgr.set_debugger(dbg)
        elif sim_dbg is not None:
            self.main_instance.debugger_list_mgr.remove_debugger(sim_dbg)

    #
    # Private methods
    #

    def _get_or_create_view(self, category: str, view_type: Type[T], position: str = "center") -> T:
        view = self.view_manager.current_view_in_category(category)
        if view is None:
            view = self.view_manager.first_view_in_category(category)
        if view is None:
            view = view_type(self, self._main_instance, position)
            self.add_view(view)
        return view



from typing import TYPE_CHECKING, Any, Optional

from angr.knowledge_plugins.key_definitions.constants import OP_AFTER, OP_BEFORE
from PySide6.QtWidgets import (
    QDialog,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QRadioButton,
    QVBoxLayout,
    QWidget,
)

from angrmanagement.utils.func import function_prototype_str

if TYPE_CHECKING:
    from angr.knowledge_plugins import Function


class DependsOn(QDialog):
    def __init__(self, addr: int, operand, instr=None, func: Optional["Function"] = None, parent=None):
        super().__init__(parent)

        self._addr = addr
        self._operand = operand
        self._instruction = instr
        self._function = func

        # output
        self.location: Optional[int] = None
        self.arg: Optional[Any] = None
        self.reg: Optional[Any] = None

        # UI widgets
        self._instr_layout: QHBoxLayout = None
        self._func_layout: QHBoxLayout = None
        self._arg_widget: QWidget = None
        self._reg_widget: QWidget = None

        self._before_radio: QRadioButton = None
        self._after_radio: QRadioButton = None
        self._arg_layout: QVBoxLayout = None
        self._arg_radiobox: QRadioButton = None
        self._arg_box: QLineEdit = None
        self._reg_layout: QVBoxLayout = None
        self._reg_radiobox: QRadioButton = None
        self._reg_box: QLineEdit = None

        self._init_widgets()

    def _init_widgets(self):
        # the instruction
        instr_lbl = QLabel("Instruction")
        instr_box = QLineEdit("TODO")
        instr_box.setReadOnly(True)

        self._instr_layout = QHBoxLayout()
        self._instr_layout.addWidget(instr_lbl)
        self._instr_layout.addWidget(instr_box)

        # the function
        func_lbl = QLabel("Function")
        func_box = QLineEdit("TODO")
        func_box.setReadOnly(True)

        self._func_layout = QHBoxLayout()
        self._func_layout.addWidget(func_lbl)
        self._func_layout.addWidget(func_box)

        #
        # location
        #
        location_group = QGroupBox("Location")

        self._before_radio = QRadioButton("Before the instruction")
        self._after_radio = QRadioButton("After the instruction")
        location_layout = QVBoxLayout()
        location_layout.addWidget(self._before_radio)
        location_layout.addWidget(self._after_radio)
        location_layout.addStretch(1)

        self._before_radio.setChecked(True)

        location_group.setLayout(location_layout)

        #
        # target atoms
        #

        # function argument
        self._arg_radiobox = QRadioButton("Function argument")
        self._arg_radiobox.clicked.connect(self._on_targetatoms_radiobutton_clicked)
        self._arg_box = QLineEdit("0")

        # register
        self._reg_radiobox = QRadioButton("Register")
        self._reg_radiobox.clicked.connect(self._on_targetatoms_radiobutton_clicked)
        self._reg_box = QLineEdit("")

        atom_type_group = QGroupBox("Atom type")
        atom_type_layout = QVBoxLayout()
        atom_type_layout.addWidget(self._arg_radiobox)
        # atom_type_layout.addWidget(self._reg_radiobox)
        atom_type_layout.addStretch(1)
        atom_type_group.setLayout(atom_type_layout)

        atom_layout = QVBoxLayout()
        arg_layout = QHBoxLayout()
        arg_lbl = QLabel("Argument index")
        self._arg_widget = QWidget()
        self._arg_widget.setLayout(arg_layout)
        arg_layout.addWidget(arg_lbl)
        arg_layout.addWidget(self._arg_box)

        reg_layout = QHBoxLayout()
        reg_lbl = QLabel("Register name")
        self._reg_widget = QWidget()
        self._reg_widget.setLayout(reg_layout)
        reg_layout.addWidget(reg_lbl)
        reg_layout.addWidget(self._reg_box)

        atom_layout.addWidget(self._arg_widget)
        atom_layout.addWidget(self._reg_widget)

        self._arg_radiobox.setChecked(True)
        self._reg_widget.setHidden(True)

        #
        # buttons
        #
        ok_btn = QPushButton("OK")
        ok_btn.clicked.connect(self._on_ok_clicked)
        cancel_btn = QPushButton("Cancel")
        cancel_btn.clicked.connect(self._on_cancel_clicked)

        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(ok_btn)
        buttons_layout.addWidget(cancel_btn)

        layout = QVBoxLayout()
        if self._function is not None:
            func_box.setText(function_prototype_str(self._function))
            layout.addLayout(self._func_layout)
        else:
            layout.addLayout(self._instr_layout)
        layout.addWidget(location_group)
        layout.addWidget(atom_type_group)
        layout.addLayout(atom_layout)
        layout.addLayout(buttons_layout)

        self.setLayout(layout)

    #
    # Events
    #

    def _targetatoms_hide_all(self):
        self._arg_widget.setHidden(True)
        self._reg_widget.setHidden(True)

    def _on_targetatoms_radiobutton_clicked(self):
        self._targetatoms_hide_all()
        if self._arg_radiobox.isChecked():
            self._arg_widget.setHidden(False)
        elif self._reg_radiobox.isChecked():
            self._reg_widget.setHidden(False)

    def _on_ok_clicked(self):
        if self._before_radio.isChecked():
            self.location = OP_BEFORE
        else:
            self.location = OP_AFTER

        if self._arg_radiobox.isChecked():
            try:
                self.arg = int(self._arg_box.text())
            except ValueError:
                # invalid argument index
                QMessageBox(self).critical(
                    self,
                    "Invalid argument index",
                    'The given function argument index "%s" is unsupported. Only integers are allowed.'
                    % self._arg_box.text(),
                    buttons=QMessageBox.Ok,
                )
                return
        else:
            raise NotImplementedError

        self.close()

    def _on_cancel_clicked(self):
        self.location = None
        self.arg = None
        self.reg = None
        self.close()



# pylint:disable=global-statement
import ctypes
import datetime
import logging
import time
from typing import Optional

from angrmanagement.logic import GlobalInfo
from angrmanagement.logic.threads import gui_thread_schedule_async

m = ...


log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


def _load_autoreload():
    """
    Load the autoreload extension module. Delay the import and initialization to reduce angr management's startup time.
    """

    global m
    try:
        from IPython.extensions.autoreload import ModuleReloader  # pylint:disable=import-outside-toplevel

        m = ModuleReloader()
        m.enabled = True
        m.check_all = True
        m.check()
    except ImportError:
        m = None


class Job:
    """
    The base class of all Jobs in angr management.
    """

    def __init__(self, name, on_finish=None, blocking=False):
        self.name = name
        self.progress_percentage = 0.0
        self.last_text: Optional[str] = None
        self.start_at: float = 0.0
        self.last_gui_updated_at: float = 0.0
        self.blocking = blocking

        # callbacks
        self._on_finish = on_finish

        if GlobalInfo.autoreload:
            if m is ...:
                _load_autoreload()
            if m is not None:
                prestate = dict(m.modules_mtimes)
                m.check()
                poststate = dict(m.modules_mtimes)
                if prestate != poststate:
                    log.warning("Auto-reload found changed modules")

    @property
    def time_elapsed(self) -> str:
        return str(datetime.timedelta(seconds=int(time.time() - self.start_at)))

    def run(self, inst):
        log.info('Job "%s" started', self.name)
        self._progress_callback(0)
        self.start_at = time.time()
        r = self._run(inst)
        now = time.time()
        duration = now - self.start_at
        log.info('Job "%s" completed after %.2f seconds', self.name, duration)
        return r

    def _run(self, inst):
        raise NotImplementedError

    def finish(self, inst, result):  # pylint: disable=unused-argument
        inst.jobs = inst.jobs[1:]

        gui_thread_schedule_async(self._finish_progress)
        if self._on_finish:
            gui_thread_schedule_async(self._on_finish)

    def keyboard_interrupt(self):
        """Called from the GUI thread when the user presses Ctrl+C or presses a cancel button"""
        # lol. lmao even.
        if GlobalInfo.main_window.workspace.main_instance.current_job == self:
            tid = GlobalInfo.main_window.workspace.main_instance.worker_thread.ident
            res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), ctypes.py_object(KeyboardInterrupt))
            if res != 1:
                ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), 0)
                log.error("Failed to interrupt thread")

    def _progress_callback(self, percentage, text=None):
        delta = percentage - self.progress_percentage

        if (delta > 0.02 or self.last_text != text) and time.time() - self.last_gui_updated_at >= 0.1:
            self.last_gui_updated_at = time.time()
            self.progress_percentage = percentage
            gui_thread_schedule_async(self._set_progress, args=(text,))

    def _set_progress(self, text=None):
        status = self.name
        if text:
            status += ": " + text
        GlobalInfo.main_window.progress(status, self.progress_percentage)

    def _finish_progress(self):
        pass



import re
from typing import TYPE_CHECKING

from angr.analyses.decompiler.structured_codegen.c import (
    CArrayTypeLength,
    CClosingObject,
    CConstant,
    CExpression,
    CFunction,
    CFunctionCall,
    CLabel,
    CStatement,
    CStructFieldNameDef,
    CVariable,
)
from angr.sim_type import SimType
from pyqodeng.core.api import SyntaxHighlighter
from PySide6.QtGui import QBrush, QColor, QFont, QTextCharFormat

from angrmanagement.config import Conf

if TYPE_CHECKING:
    from angrmanagement.ui.documents import QCodeDocument

FORMATS = {}


def create_char_format(color: QColor, weight: QFont.Weight, style: QFont.Style) -> QTextCharFormat:
    f = QTextCharFormat()
    f.setForeground(QBrush(color))
    f.setFontWeight(weight)
    if style == QFont.Style.StyleItalic:
        f.setFontItalic(True)
    return f


def reset_formats():
    FORMATS["keyword"] = create_char_format(
        Conf.pseudocode_keyword_color, Conf.pseudocode_keyword_weight, Conf.pseudocode_keyword_style
    )

    FORMATS["quotation"] = create_char_format(
        Conf.pseudocode_quotation_color, Conf.pseudocode_quotation_weight, Conf.pseudocode_quotation_style
    )

    FORMATS["function"] = create_char_format(
        Conf.pseudocode_function_color, Conf.pseudocode_function_weight, Conf.pseudocode_function_style
    )

    FORMATS["comment"] = create_char_format(
        Conf.pseudocode_comment_color, Conf.pseudocode_comment_weight, Conf.pseudocode_comment_style
    )

    FORMATS["variable"] = create_char_format(
        Conf.pseudocode_variable_color, Conf.pseudocode_variable_weight, Conf.pseudocode_variable_style
    )

    FORMATS["type"] = create_char_format(
        Conf.pseudocode_types_color, Conf.pseudocode_types_weight, Conf.pseudocode_types_style
    )

    FORMATS["label"] = create_char_format(
        Conf.pseudocode_label_color, Conf.pseudocode_label_weight, Conf.pseudocode_label_style
    )


def _format_node(obj):
    """
    Return the format for the given node.
    """
    if isinstance(obj, SimType):
        return FORMATS["type"]
    elif isinstance(obj, (CFunctionCall, CFunction)):
        return FORMATS["function"]
    elif isinstance(obj, CLabel):
        return FORMATS["label"]
    elif isinstance(obj, CVariable):
        return FORMATS["variable"]
    elif isinstance(obj, CArrayTypeLength):
        # This is the part that goes after a fixed-size array (the
        # "[20]" in "char foo[20];"), and it's highly unlikely
        # that anyone will want to change the color here. But if
        # you do, follow the format of the rest.
        return None
    elif isinstance(obj, CStructFieldNameDef):
        # This is the part that is a field name in a struct def,
        # and it's highly unlikely that anyone will want to change
        # the color here. But if you do, follow the format of the
        # rest.
        return None
    elif isinstance(obj, (CClosingObject, CStatement, CConstant, CExpression)):
        return None
    else:
        return None


reset_formats()


class QCCodeHighlighter(SyntaxHighlighter):
    """
    A syntax highlighter for QCCodeEdit. Uses a custom lexing scheme to detect C constructs (functions, keywords,
    comments, and strings) and adds styling to them based on the current color scheme.
    """

    HIGHLIGHTING_RULES = [
        # quotation
        # (r"\"([^\\\"]|(\\.))*\"", 'quotation'),
        # comment
        # (r"//[^\n]*", 'comment'),
        # (r"/\*[^\n]*\*/", 'comment'),
        # function
        (r"\bNULL\b", "function"),
        # keywords
        (r"\bauto\b", "keyword"),
        (r"\bbreak\b", "keyword"),
        (r"\bcase\b", "keyword"),
        (r"\bcatch\b", "keyword"),
        (r"\bclass\b", "keyword"),
        (r"\bconst\b", "keyword"),
        (r"\bcontinue\b", "keyword"),
        (r"\bdefault\b", "keyword"),
        (r"\bdelete\b", "keyword"),
        (r"\bdo\b", "keyword"),
        (r"\belse\b", "keyword"),
        (r"\benum\b", "keyword"),
        (r"\bexplicit\b", "keyword"),
        (r"\bextern\b", "keyword"),
        (
            r"\bfalse\b",
            "keyword",
        ),  # TODO: false isn't really a keyword, and other highlighters style it different from keywords
        (r"\bfor\b", "keyword"),
        (r"\bfriend\b", "keyword"),
        (r"\bgoto\b", "keyword"),
        (r"\bif\b", "keyword"),
        (r"\binline\b", "keyword"),
        (r"\bnamespace\b", "keyword"),
        (r"\bnew\b", "keyword"),
        (r"\boperator\b", "keyword"),
        (r"\bprivate\b", "keyword"),
        (r"\bprotected\b", "keyword"),
        (r"\bpublic\b", "keyword"),
        (r"\bregister\b", "keyword"),
        (r"\breturn\b", "keyword"),
        (r"\bsizeof\b", "keyword"),
        (r"\bstatic\b", "keyword"),
        (r"\bstruct\b", "keyword"),
        (r"\bswitch\b", "keyword"),
        (r"\btemplate\b", "keyword"),
        (r"\bthis\b", "keyword"),
        (
            r"\btrue\b",
            "keyword",
        ),  # TODO: true isn't really a keyword, and other highlighters style it different from keywords
        (r"\btypedef\b", "keyword"),
        (r"\btypename\b", "keyword"),
        (r"\bunion\b", "keyword"),
        (r"\bvirtual\b", "keyword"),
        (r"\bvolatile\b", "keyword"),
        (r"\bwhile\b", "keyword"),
    ]

    def __init__(self, parent, color_scheme=None):
        super().__init__(parent, color_scheme=color_scheme)

        self.doc: QCodeDocument = parent
        self.comment_status = False

    def highlight_block(self, text, block):
        # this code makes the assumption that this function is only ever called on lines in sequence in order
        # it might also fuck up if it ever calls it starting in the middle...
        if block.previous() is None:
            self.comment_status = False

        mark_in = 0

        quote_status = False
        quote_mark = None
        escape_counter = 0
        for col, _ in enumerate(text):
            if quote_status:
                assert not self.comment_status
                if escape_counter:
                    escape_counter -= 1
                elif text[col] == "\\":
                    escape_counter = 1
                elif text[col] == quote_mark:
                    quote_status = False
                    mark_out = col + 1
                    self.setFormat(mark_in, mark_out - mark_in, FORMATS["quotation"])
                    text = text[:mark_in] + " " * (mark_out - mark_in) + text[mark_out:]
            else:
                if self.comment_status:
                    if text[col : col + 2] == "*/":
                        mark_out = col + 2
                        self.comment_status = False
                        self.setFormat(mark_in, mark_out - mark_in, FORMATS["comment"])
                        text = text[:mark_in] + " " * (mark_out - mark_in) + text[mark_out:]
                else:
                    # regular mode is here
                    if text[col] in ('"', "'"):
                        mark_in = col
                        quote_status = True
                        quote_mark = text[col]
                    elif text[col : col + 2] == "/*":
                        mark_in = col
                        self.comment_status = True
                    elif text[col : col + 2] == "//":
                        # do not set comment_status. just format the line and break.
                        mark_in = col
                        mark_out = len(text)
                        self.setFormat(mark_in, mark_out - mark_in, FORMATS["comment"])
                        text = text[:mark_in] + " " * (mark_out - mark_in) + text[mark_out:]
                        break
        if self.comment_status:
            # do not unset. this is a multiline comment
            mark_out = len(text)
            self.setFormat(mark_in, mark_out - mark_in, FORMATS["comment"])
            text = text[:mark_in] + " " * (mark_out - mark_in) + text[mark_out:]

        # Go through AST and see about marking styles
        start_pos = block.position()
        current_idx = 0
        while current_idx < len(text):
            # if we know it's whitespace (b/c that can happen above due to comments), just skip it
            if text[current_idx] == " ":
                current_idx += 1
                continue

            current_pos = start_pos + current_idx
            element = self.doc.posmap.get_element(current_pos)
            if not element:
                current_idx += 1
                continue

            fmt = _format_node(element.obj)

            # Because of skipping spaces, we might end up inside an element, so don't use current_idx in the following
            if fmt:
                self.setFormat(element.start - start_pos, element.length, fmt)
            current_idx = (element.start - start_pos) + element.length

        for pattern, format_id in self.HIGHLIGHTING_RULES:
            for mo in list(re.finditer(pattern, text)):
                start = mo.start()
                end = mo.end()
                self.setFormat(start, end - start, FORMATS[format_id])
                # if format_id in {'quotation', 'comment'}:
                #    # remove the formatted parts so that we do not end up highlighting these parts again
                #    text = text[:start] + " " * (end-start) + text[end:]



import logging
from typing import TYPE_CHECKING, Optional

from angr.analyses.decompiler.utils import to_ail_supergraph
from PySide6.QtCore import QEvent, QPointF, QRect, QRectF, QSize, Qt, QTimeLine
from PySide6.QtWidgets import QFrame

from angrmanagement.config import Conf
from angrmanagement.utils import get_out_branches
from angrmanagement.utils.cfg import categorize_edges
from angrmanagement.utils.graph_layouter import GraphLayouter

from .qblock import QGraphBlock
from .qdisasm_base_control import DisassemblyLevel, QDisassemblyBaseControl
from .qgraph import QZoomableDraggableGraphicsView
from .qgraph_arrow import QDisasmGraphArrow
from .qminimap import QMiniMapView

if TYPE_CHECKING:
    from angrmanagement.logic.disassembly import InfoDock


_l = logging.getLogger(__name__)


class QViewPortMover:
    def __init__(
        self,
        disasm_graph: "QDisassemblyGraph",
        x: int,
        y: int,
        target_x: int,
        target_y: int,
        interval: int = 700,
        max_frame: int = 100,
    ):
        self.disasm_graph = disasm_graph
        self.target_x = target_x
        self.target_y = target_y

        self.initial_x = x
        self.initial_y = y
        self.x_step = (self.target_x - self.initial_x) / max_frame
        self.y_step = (self.target_y - self.initial_y) / max_frame

        self._move_timeline = QTimeLine(interval)
        self._move_timeline.setFrameRange(0, max_frame)
        self._move_timeline.setUpdateInterval(10)

    def start(self):
        self._move_timeline.frameChanged.connect(self._set_pos)
        self._move_timeline.start()

    def _set_pos(self, step):
        self.disasm_graph.centerOn(self.initial_x + self.x_step * step, self.initial_y + self.y_step * step)


class QDisassemblyGraph(QDisassemblyBaseControl, QZoomableDraggableGraphicsView):
    def __init__(self, instance, disasm_view, parent=None):
        QDisassemblyBaseControl.__init__(self, instance, disasm_view, QZoomableDraggableGraphicsView)
        QZoomableDraggableGraphicsView.__init__(self, parent=parent)

        self.instance = instance

        self.disasm = None
        self.variable_manager = None

        self._function_graph = None
        self._supergraph = None
        self._viewport_mover = None

        self._edges = None
        self._arrows = []  # A list of references to QGraphArrow objects

        self.blocks = []

        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.setFrameStyle(QFrame.NoFrame)
        self.setBackgroundBrush(Conf.disasm_view_background_color)

        self._minimap = QMiniMapView(self, parent=self)
        self._minimap.setMaximumSize(200, 400)
        self._minimap.move(20, 20)
        self._minimap.setVisible(self.disasm_view.show_minimap)

    #
    # Properties
    #

    @property
    def function_graph(self):
        return self._function_graph

    @function_graph.setter
    def function_graph(self, v):
        if v is not self._function_graph:
            self._function_graph = v

            self.reload()

    @property
    def induction_variable_analysis(self):
        return self.infodock.induction_variable_analysis

    @induction_variable_analysis.setter
    def induction_variable_analysis(self, v):
        self.infodock.induction_variable_analysis = v

    #
    # Public methods
    #

    def reload(self, old_infodock: Optional["InfoDock"] = None):
        # if there is an instruction in selection, we will want to select that instruction again after reloading this
        # view.
        selected_insns = old_infodock.selected_insns.am_obj if old_infodock is not None else set()

        self._reset_scene()
        self._arrows.clear()
        self._minimap.reload_target_scene()
        self.blocks.clear()
        self._insaddr_to_block.clear()
        if self._function_graph is None:
            return

        scene = self.scene()

        if self._disassembly_level is DisassemblyLevel.AIL:
            func = self._function_graph.function
            try:
                # always check if decompiler has cached a clinic object first
                self.disasm = self.instance.kb.structured_code[(func.addr, "pseudocode")].clinic
            except (KeyError, AttributeError):
                self.disasm = self.instance.project.analyses.Clinic(func)

            self._supergraph = to_ail_supergraph(self.disasm.cc_graph)

            def nodefunc(n):
                return n

            def branchfunc(n):
                return None

            has_idx = True
        else:
            include_ir = self._disassembly_level is DisassemblyLevel.LifterIR
            self.disasm = self.instance.project.analyses.Disassembly(
                function=self._function_graph.function, include_ir=include_ir
            )
            view = self.disasm_view.workspace.view_manager.first_view_in_category("console")
            if view is not None:
                view.push_namespace(
                    {
                        "disasm": self.disasm,
                    }
                )
            self._supergraph = self._function_graph.supergraph

            def nodefunc(n):
                return n.cfg_nodes

            branchfunc = get_out_branches
            has_idx = False

        for n in self._supergraph.nodes():
            block = QGraphBlock(
                self.instance,
                self._function_graph.function.addr,
                self.disasm_view,
                self.disasm,
                self.infodock,
                n.addr,
                nodefunc(n),
                branchfunc(n),
                scene,
                idx=n.idx if has_idx else None,
            )
            if n.addr == self._function_graph.function.addr:
                self.entry_block = block
            scene.addItem(block)
            self.blocks.append(block)

            for insn_addr in block.addr_to_insns:
                self._insaddr_to_block[insn_addr] = block

        self.request_relayout()
        self._update_scene_boundary()

        # determine initial view focus point
        self._reset_view()

        # select the old instructions
        for insn_addr in selected_insns:
            self.infodock.select_instruction(insn_addr, unique=False, use_animation=False)

        self._minimap.reload_target_scene()

    def refresh(self):
        if not self.blocks:
            return

        for b in self.blocks:
            b.clear_cache()
            b.refresh()

        self.request_relayout()
        self._update_scene_boundary()

        self._minimap.reload_target_scene()
        self._minimap.setVisible(self.disasm_view.show_minimap)

    def set_extra_render_pass(self, is_extra_pass: bool):
        super().set_extra_render_pass(is_extra_pass)
        if not is_extra_pass:
            # We hide block objects in low LoD passes. Restore them now if
            # they were hidden.
            for b in self.blocks:
                b.restore_temporarily_hidden_objects()

    #
    # Event handlers
    #

    def event(self, event):
        """
        Reimplemented to capture the Tab keypress event.
        """

        # by default, the tab key moves focus. Hijack the tab key
        if event.type() == QEvent.KeyPress and event.key() == Qt.Key_Tab:
            self.disasm_view.keyPressEvent(event)
            return True
        return super().event(event)

    def mousePressEvent(self, event):
        btn = event.button()

        if btn == Qt.ForwardButton:
            self.disasm_view.jump_forward()
        elif btn == Qt.BackButton:
            self.disasm_view.jump_back()
        else:
            super().mousePressEvent(event)

    def changeEvent(self, event: QEvent):
        """
        Redraw on color scheme update.
        """
        if event.type() == QEvent.PaletteChange:
            self.setBackgroundBrush(Conf.disasm_view_background_color)
            self.reload()

    def on_background_click(self):
        pass

    #
    # Layout
    #

    def _graph_size(self):
        width, height = 0, 0

        for block in self.blocks:
            if block.x + block.width > width:
                width = block.x + block.width
            if block.y + block.height > height:
                height = block.y + block.height

        # TODO: Check all edges as well

        return QSize(width, height)

    def _layout_graph(self):
        node_sizes = {}
        node_map = {}
        for block in self.blocks:
            node_map[block.addr] = block
        for node in self._supergraph.nodes():
            block = node_map[node.addr]
            node_sizes[node] = block.width, block.height
        gl = GraphLayouter(self._supergraph, node_sizes)

        nodes = {}
        for node, coords in gl.node_coordinates.items():
            nodes[(node.addr, node.idx)] = coords

        return nodes, gl.edges

    def request_relayout(self):
        node_coords, edges = self._layout_graph()

        self._edges = edges

        categorize_edges(self.disasm, edges)

        if not node_coords:
            print("Failed to get node_coords")
            return

        # layout nodes
        for block in self.blocks:
            x, y = node_coords[(block.addr, block.idx)]
            block.setPos(x, y)

        scene = self.scene()

        # remove exiting arrows
        for arrow in self._arrows:
            scene.removeItem(arrow)
        self._arrows.clear()

        for edge in self._edges:
            arrow = QDisasmGraphArrow(edge, self.disasm_view, self.infodock)
            self._arrows.append(arrow)
            scene.addItem(arrow)
            arrow.setPos(QPointF(*edge.coordinates[0]))

    def _update_scene_boundary(self):
        scene = self.scene()
        # Leave some margins
        rect: QRectF = scene.itemsBoundingRect()
        scene.setSceneRect(QRectF(rect.x() - 200, rect.y() - 200, rect.width() + 400, rect.height() + 400))

    def show_instruction(self, insn_addr, insn_pos=None, centering=False, use_block_pos=False, use_animation=True):
        block: QGraphBlock = self._insaddr_to_block.get(insn_addr, None)
        if block is not None:
            if use_block_pos:
                x, y = block.mapToScene(block.x(), block.y())
            else:
                pos = block.mapToScene(*block.instruction_position(insn_addr))
                x, y = pos.x(), pos.y()

            if not centering:
                # is it visible?
                viewport = self.viewport()
                visible_area = self.mapToScene(QRect(0, 0, viewport.width(), viewport.height())).boundingRect()
                topx = visible_area.x()
                topy = visible_area.y()
                if topx <= x < topx + visible_area.width() and topy <= y < topy + visible_area.height():
                    return

            # make it visible in the center
            if use_animation:
                viewport = self.viewport()
                current_pos = self.mapToScene(viewport.width() // 2, viewport.height() // 2)
                self._viewport_mover = QViewPortMover(self, current_pos.x(), current_pos.y(), x, y)
                self._viewport_mover.start()
            else:
                self.centerOn(x, y)

    def update_label(self, addr, is_renaming=False):
        block: QGraphBlock = self._insaddr_to_block.get(addr, None)
        if block is not None:
            if is_renaming:
                # we just need to refresh the current block
                # block.refresh()  # TODO: We should be able to just refresh that single label
                block.reload()
            else:
                # life is hard... we need to reload the block, and then re-layout the entire graph because of size
                # changes
                block.reload()
                self.request_relayout()

    #
    # Private methods
    #

    def _initial_position(self):
        entry_block_rect = self.entry_block.mapRectToScene(self.entry_block.boundingRect())
        viewport_height = self.viewport().rect().height()
        min_rect = self.scene().itemsBoundingRect()
        if min_rect.height() < (viewport_height // 1.5):
            return min_rect.center()
        else:
            focus_point = (entry_block_rect.center().x(), entry_block_rect.top() + (viewport_height // 4))
            return QPointF(*focus_point)



from angrmanagement.logic import GlobalInfo

from .menu import Menu


class FunctionContextMenu(Menu):
    def __init__(self, parent):
        super().__init__("Function", parent=parent)

        self.funcs = []

        # TODO add Rename, Change Type, xrefs, etc

    def set(self, funcs):
        self.funcs = funcs
        return self

    def qmenu(self, extra_entries=None):
        if extra_entries is None:
            extra_entries = ()
        return super().qmenu(
            extra_entries=list(GlobalInfo.main_window.workspace.plugins.build_context_menu_functions(self.funcs))
            + list(extra_entries)
        )



import itertools
from collections import defaultdict

import networkx
from angr.knowledge_plugins import Function


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    args = [iter(iterable)] * n
    return itertools.izip_longest(*args, fillvalue=fillvalue)


def to_supergraph(transition_graph):
    """
    Convert transition graph of a function to a super transition graph. A super transition graph is a graph that looks
    like IDA Pro's CFG, where calls to returning functions do not terminate basic blocks.

    :param networkx.DiGraph transition_graph: The transition graph.
    :return: A converted super transition graph
    :rtype networkx.DiGraph
    """

    # make a copy of the graph
    transition_graph = networkx.DiGraph(transition_graph)

    # remove all edges that transitions to outside
    for src, dst, data in list(transition_graph.edges(data=True)):
        if data["type"] in ("transition", "exception") and data.get("outside", False) is True:
            transition_graph.remove_edge(src, dst)
        if transition_graph.in_degree(dst) == 0:
            transition_graph.remove_node(dst)

    edges_to_shrink = set()

    # Find all edges to remove in the super graph
    for src in transition_graph.nodes():
        edges = transition_graph[src]

        # there are two types of edges we want to remove:
        # - call or fakerets, since we do not want blocks to break at calls
        # - boring jumps that directly transfer the control to the block immediately after the current block. this is
        #   usually caused by how VEX breaks down basic blocks, which happens very often in MIPS

        if len(edges) == 1 and src.addr + src.size == next(iter(edges.keys())).addr:
            dst = next(iter(edges.keys()))
            dst_in_edges = transition_graph.in_edges(dst)
            if len(dst_in_edges) == 1:
                edges_to_shrink.add((src, dst))
                continue

        if any(iter("type" in data and data["type"] not in ("fake_return", "call") for data in edges.values())):
            continue

        for dst, data in edges.items():
            if isinstance(dst, Function):
                continue
            if "type" in data and data["type"] == "fake_return":
                if all(
                    iter(
                        "type" in data and data["type"] in ("fake_return", "return")
                        for _, _, data in transition_graph.in_edges(dst, data=True)
                    )
                ):
                    edges_to_shrink.add((src, dst))
                break

    # Create the super graph
    super_graph = networkx.DiGraph()

    supernodes_map = {}

    function_nodes = set()  # it will be traversed after all other nodes are added into the supergraph

    for node in transition_graph.nodes():
        if isinstance(node, Function):
            function_nodes.add(node)
            # don't put functions into the supergraph
            continue

        dests_and_data = transition_graph[node]

        # make a super node
        if node in supernodes_map:
            src_supernode = supernodes_map[node]
        else:
            src_supernode = SuperCFGNode.from_cfgnode(node)
            supernodes_map[node] = src_supernode
            # insert it into the graph
            super_graph.add_node(src_supernode)

        if not dests_and_data:
            # might be an isolated node
            continue

        # Take src_supernode off the graph since we might modify it
        if src_supernode in super_graph:
            existing_in_edges = list(super_graph.in_edges(src_supernode, data=True))
            existing_out_edges = list(super_graph.out_edges(src_supernode, data=True))
            super_graph.remove_node(src_supernode)
        else:
            existing_in_edges = []
            existing_out_edges = []

        for dst, data in dests_and_data.items():
            edge = (node, dst)

            if edge in edges_to_shrink:
                dst_supernode = supernodes_map[dst] if dst in supernodes_map else None

                src_supernode.insert_cfgnode(dst)

                # update supernodes map
                supernodes_map[dst] = src_supernode

                # merge the other supernode
                if dst_supernode is not None:
                    src_supernode.merge(dst_supernode)

                    for src in dst_supernode.cfg_nodes:
                        supernodes_map[src] = src_supernode

                    # link all out edges of dst_supernode to src_supernode
                    for dst_, data_ in super_graph[dst_supernode].items():
                        super_graph.add_edge(src_supernode, dst_, **data_)

                    # link all in edges of dst_supernode to src_supernode
                    for src_, _, data_ in super_graph.in_edges(dst_supernode, data=True):
                        super_graph.add_edge(src_, src_supernode, **data_)

                        if "type" in data_ and data_["type"] in {"transition", "exception", "call"}:
                            if not ("ins_addr" in data_ and "stmt_idx" in data_):
                                # this is a hack to work around the issue in Function.normalize() where ins_addr and
                                # stmt_idx weren't properly set onto edges
                                continue
                            src_supernode.register_out_branch(
                                data_["ins_addr"], data_["stmt_idx"], data_["type"], dst_supernode.addr
                            )

                    super_graph.remove_node(dst_supernode)

            else:
                if isinstance(dst, Function):
                    # skip all functions
                    continue

                # make a super node
                if dst in supernodes_map:
                    dst_supernode = supernodes_map[dst]
                else:
                    dst_supernode = SuperCFGNode.from_cfgnode(dst)
                    supernodes_map[dst] = dst_supernode

                super_graph.add_edge(src_supernode, dst_supernode, **data)

                if "type" in data and data["type"] in {"transition", "exception", "call"}:
                    if not ("ins_addr" in data and "stmt_idx" in data):
                        # this is a hack to work around the issue in Function.normalize() where ins_addr and
                        # stmt_idx weren't properly set onto edges
                        continue
                    src_supernode.register_out_branch(
                        data["ins_addr"], data["stmt_idx"], data["type"], dst_supernode.addr
                    )

        # add back the node (in case there are no edges)
        super_graph.add_node(src_supernode)
        # add back the old edges
        for src, _, data in existing_in_edges:
            super_graph.add_edge(src, src_supernode, **data)
        for _, dst, data in existing_out_edges:
            super_graph.add_edge(src_supernode, dst, **data)

    for node in function_nodes:
        in_edges = transition_graph.in_edges(node, data=True)

        for src, _, data in in_edges:
            if not ("ins_addr" in data and "stmt_idx" in data):
                # this is a hack to work around the issue in Function.normalize() where ins_addr and
                # stmt_idx weren't properly set onto edges
                continue
            supernode = supernodes_map[src]
            supernode.register_out_branch(data["ins_addr"], data["stmt_idx"], data["type"], node.addr)

    return super_graph


class OutBranch:
    def __init__(self, ins_addr, stmt_idx, branch_type):
        self.ins_addr = ins_addr
        self.stmt_idx = stmt_idx
        self.type = branch_type

        self.targets = set()

    def __repr__(self):
        if self.ins_addr is None:
            return "<OutBranch at None, type %s>" % self.type
        return f"<OutBranch at {self.ins_addr:#x}, type {self.type}>"

    def add_target(self, addr):
        self.targets.add(addr)

    def merge(self, other):
        """
        Merge with the other OutBranch descriptor.

        :param OutBranch other: The other item to merge with.
        :return: None
        """

        assert self.ins_addr == other.ins_addr
        assert self.type == other.type

        o = self.copy()
        o.targets |= other.targets

        return o

    def copy(self):
        o = OutBranch(self.ins_addr, self.stmt_idx, self.type)
        o.targets = self.targets.copy()
        return o

    def __eq__(self, other):
        if not isinstance(other, OutBranch):
            return False

        return (
            self.ins_addr == other.ins_addr
            and self.stmt_idx == other.stmt_idx
            and self.type == other.type
            and self.targets == other.targets
        )

    def __hash__(self):
        return hash((self.ins_addr, self.stmt_idx, self.type))


class SuperCFGNode:
    def __init__(self, addr, idx=None):
        self.addr = addr
        self.idx = idx

        self.cfg_nodes = []

        self.out_branches = defaultdict(dict)

    @property
    def size(self):
        return sum(node.size for node in self.cfg_nodes)

    @classmethod
    def from_cfgnode(cls, cfg_node):
        s = cls(cfg_node.addr)

        s.cfg_nodes.append(cfg_node)

        return s

    def insert_cfgnode(self, cfg_node):
        # TODO: Make it binary search/insertion
        for i, n in enumerate(self.cfg_nodes):
            if cfg_node.addr < n.addr:
                # insert before n
                self.cfg_nodes.insert(i, cfg_node)
                break
            elif cfg_node.addr == n.addr:
                break
        else:
            self.cfg_nodes.append(cfg_node)

        # update addr
        self.addr = self.cfg_nodes[0].addr

    def register_out_branch(self, ins_addr, stmt_idx, branch_type, target_addr):
        if ins_addr not in self.out_branches or stmt_idx not in self.out_branches[ins_addr]:
            self.out_branches[ins_addr][stmt_idx] = OutBranch(ins_addr, stmt_idx, branch_type)

        self.out_branches[ins_addr][stmt_idx].add_target(target_addr)

    def merge(self, other):
        """
        Merge another supernode into the current one.

        :param SuperCFGNode other: The supernode to merge with.
        :return: None
        """

        for n in other.cfg_nodes:
            self.insert_cfgnode(n)

        for ins_addr, outs in other.out_branches.items():
            if ins_addr in self.out_branches:
                for stmt_idx, item in outs.items():
                    if stmt_idx in self.out_branches[ins_addr]:
                        self.out_branches[ins_addr][stmt_idx].merge(item)
                    else:
                        self.out_branches[ins_addr][stmt_idx] = item

            else:
                item = next(iter(outs.values()))
                self.out_branches[ins_addr][item.stmt_idx] = item

    def __repr__(self):
        return "<SuperCFGNode %#08x, %d blocks, %d out branches>" % (
            self.addr,
            len(self.cfg_nodes),
            len(self.out_branches),
        )

    def __hash__(self):
        return hash(("supercfgnode", self.addr))

    def __eq__(self, other):
        if not isinstance(other, SuperCFGNode):
            return False

        return self.addr == other.addr



from typing import TYPE_CHECKING

from PySide6.QtCore import QSize
from PySide6.QtWidgets import QComboBox, QHBoxLayout, QLabel, QLineEdit, QPushButton, QVBoxLayout

from angrmanagement.ui.views.view import BaseView

from .qsearch_table import QSearchTable

if TYPE_CHECKING:
    from angr.knowledge_plugins.cfg.memory_data import MemoryData


class SearchView(BaseView):
    """
    Container view for the QSearchTable Object.
    Has handlers for switching between search types and executing the search.
    """

    def __init__(self, plugin, workspace, instance, default_docking_position, *args, **kwargs):
        super().__init__("search", workspace, instance, default_docking_position, *args, **kwargs)

        self.base_caption = "Search"
        self.plugin = plugin

        self._search_table: QSearchTable
        self._type_list: QComboBox

        self._selected_type = "bytes"

        self._init_widgets()
        self.reload()

    def reload(self):
        pass

    def sizeHint(self):
        return QSize(400, 800)

    #
    # Event handlers
    #

    def _on_search_click(self):
        self._selected_type = self._type_list.currentText()
        pattern = self._filter_string.text()
        self._value_table.filter_string = pattern
        self.reload()

    def _on_string_selected(self, s: "MemoryData"):
        """
        A string reference is selected.

        :param s:
        :return:
        """

        if len(self.workspace.view_manager.views_by_category["hex"]) == 1:
            hex_view = self.workspace.view_manager.first_view_in_category("hex")
        else:
            hex_view = self.workspace.view_manager.current_view_in_category("hex")
        if hex_view is not None:
            hex_view.jump_to(s[0])
            self.workspace.view_manager.raise_view(hex_view)

    #
    # Private methods
    #

    def _init_widgets(self):
        self._type_list = QComboBox(parent=self)
        self._type_list.addItems(["bytes", "int", "float"])
        self._search_button = QPushButton("Search", parent=self)
        self._filter_string = QLineEdit(self)
        self._search_button.clicked.connect(self._on_search_click)

        search_layout = QHBoxLayout()
        search_layout.addWidget(QLabel("Type:", self))
        search_layout.addWidget(self._type_list, 10)
        search_layout.addWidget(QLabel("Query:", self))
        search_layout.addWidget(self._filter_string, 10)
        search_layout.addWidget(self._search_button)
        search_layout.setContentsMargins(3, 3, 3, 3)
        search_layout.setSpacing(3)

        self._value_table = QSearchTable(self.instance, self, selection_callback=self._on_string_selected)

        layout = QVBoxLayout()
        layout.addLayout(search_layout)
        layout.addWidget(self._value_table)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)



import enum
from datetime import datetime

from bidict import bidict
from PySide6.QtGui import QColor
from PySide6.QtWidgets import (
    QAbstractScrollArea,
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFrame,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListView,
    QListWidget,
    QListWidgetItem,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)

from angrmanagement.config import Conf, save_config
from angrmanagement.config.color_schemes import COLOR_SCHEMES
from angrmanagement.config.config_manager import ENTRIES
from angrmanagement.logic.url_scheme import AngrUrlScheme
from angrmanagement.ui.css import refresh_theme
from angrmanagement.ui.widgets.qcolor_option import QColorOption
from angrmanagement.ui.widgets.qfont_option import QFontOption


class Page(QWidget):
    """
    Base class for pages.
    """

    def save_config(self):
        raise NotImplementedError

    NAME = NotImplemented


class Integration(Page):
    """
    The integration page.
    """

    NAME = "OS Integration"

    def __init__(self, parent=None):
        super().__init__(parent)

        self._url_scheme_chk: QCheckBox
        self._url_scheme_text: QLineEdit

        self._init_widgets()
        self._load_config()

    def _init_widgets(self):
        # os integration
        os_integration = QGroupBox("OS integration")
        self._url_scheme_chk = QCheckBox("Register angr URL scheme (angr://).")
        self._url_scheme_text = QLineEdit()
        self._url_scheme_text.setReadOnly(True)
        url_scheme_lbl = QLabel("Currently registered to:")

        os_layout = QVBoxLayout()
        os_layout.addWidget(self._url_scheme_chk)
        os_layout.addWidget(url_scheme_lbl)
        os_layout.addWidget(self._url_scheme_text)

        os_integration.setLayout(os_layout)

        layout = QVBoxLayout()
        layout.addWidget(os_integration)
        layout.addStretch()
        self.setLayout(layout)

    def _load_config(self):
        scheme = AngrUrlScheme()
        try:
            registered, register_as = scheme.is_url_scheme_registered()
            self._url_scheme_chk.setChecked(registered)
            self._url_scheme_text.setText(register_as)
        except NotImplementedError:
            # the current OS is not supported
            self._url_scheme_chk.setDisabled(True)

    def save_config(self):
        scheme = AngrUrlScheme()
        try:
            registered, _ = scheme.is_url_scheme_registered()
            if registered != self._url_scheme_chk.isChecked():
                # we need to do something
                if self._url_scheme_chk.isChecked():
                    scheme.register_url_scheme()
                else:
                    scheme.unregister_url_scheme()
        except NotImplementedError:
            # the current OS is not supported
            pass


class ThemeAndColors(Page):
    """
    Theme and Colors preferences page.
    """

    NAME = "Theme and Colors"

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self._colors_to_save = {}
        self._conf_to_save = {}
        self._schemes_combo: QComboBox = None

        self._init_widgets()

    def _init_widgets(self):
        page_layout = QVBoxLayout()

        scheme_loader_layout = QHBoxLayout()
        color_scheme_lbl = QLabel("Load Theme:")
        color_scheme_lbl.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        scheme_loader_layout.addWidget(color_scheme_lbl)

        self._schemes_combo = QComboBox(self)
        current_theme_idx = 0
        for idx, name in enumerate(["Current"] + sorted(COLOR_SCHEMES)):
            if name == Conf.theme_name:
                current_theme_idx = idx
            self._schemes_combo.addItem(name)
        self._schemes_combo.setCurrentIndex(current_theme_idx)
        scheme_loader_layout.addWidget(self._schemes_combo)
        load_btn = QPushButton("Load")
        load_btn.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        load_btn.clicked.connect(self._on_load_scheme_clicked)
        scheme_loader_layout.addWidget(load_btn)
        page_layout.addLayout(scheme_loader_layout)

        edit_colors_layout = QVBoxLayout()
        for ce in ENTRIES:
            if ce.type_ is QColor:
                row = QColorOption(getattr(Conf, ce.name), ce.name)
                edit_colors_layout.addWidget(row)
                self._colors_to_save[ce.name] = (ce, row)
            elif issubclass(ce.type_, enum.Enum):
                self._conf_to_save[ce.name] = ce.value

        frame = QFrame()
        frame.setLayout(edit_colors_layout)

        scroll = QScrollArea()
        scroll.setWidget(frame)

        scroll_layout = QHBoxLayout()
        scroll_layout.addWidget(scroll)

        page_layout.addLayout(scroll_layout)

        self.setLayout(page_layout)

    def _load_color_scheme(self, name):
        for prop, value in COLOR_SCHEMES[name].items():
            if prop in self._colors_to_save:
                row = self._colors_to_save[prop][1]
                row.set_color(value)
            if prop in self._conf_to_save:
                self._conf_to_save[prop] = value

    def _on_load_scheme_clicked(self):
        self._load_color_scheme(self._schemes_combo.currentText())
        self.save_config()

    def save_config(self):
        # pylint: disable=assigning-non-slot
        Conf.theme_name = self._schemes_combo.currentText()
        for ce, row in self._colors_to_save.values():
            setattr(Conf, ce.name, row.color.am_obj)
        for name, value in self._conf_to_save.items():
            setattr(Conf, name, value)


class Style(Page):
    """
    Preference pane for UI style choices
    """

    NAME = "Style"

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._init_widgets()

    def _init_widgets(self):
        page_layout = QVBoxLayout(self)

        # Log format
        log_format_layout = QHBoxLayout()
        log_format_lbl = QLabel("Log datetime Format String:")
        log_format_lbl.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        log_format_layout.addWidget(log_format_lbl)

        self.log_format_entry = QComboBox(self)
        fmt: str = Conf.log_timestamp_format
        ts = datetime.now()
        self._fmt_map = bidict({ts.strftime(i): i for i in [fmt, "%X", "%c"]})
        # fmt must be in _fmt_map.inverse for this to work
        if fmt not in self._fmt_map.inverse:
            fmt = self._fmt_map[ts.strftime(fmt)]
        for i in self._fmt_map:
            self.log_format_entry.addItem(i)
        # pylint: disable=unsubscriptable-object
        self.log_format_entry.setCurrentText(self._fmt_map.inverse[fmt])
        self.log_format_entry.setEditable(True)
        log_format_layout.addWidget(self.log_format_entry)
        page_layout.addLayout(log_format_layout)

        # Font options
        self._font_options = [
            QFontOption("Application Font", "ui_default_font", self),
            # TODO: other app fonts, things which set them respect updates to them in Conf
            # QFontOption("Tab View Font", "tabular_view_font", self),
            # QFontOption("Disassembly Font", "disasm_font", self),
            # QFontOption("SymExc Font", "symexec_font", self),
            # QFontOption("Code Font", "code_font", self),
        ]
        font_layout = QVBoxLayout()
        for i in self._font_options:
            font_layout.addWidget(i)
        page_layout.addLayout(font_layout)

        page_layout.addStretch()

    def save_config(self):
        fmt = self.log_format_entry.currentText()
        if fmt:
            Conf.log_timestamp_format = self._fmt_map.get(fmt, fmt)
        for i in self._font_options:
            i.update()


class Preferences(QDialog):
    """
    Application preferences dialog.
    """

    def __init__(self, workspace, parent=None):
        super().__init__(parent)

        self.workspace = workspace

        self._pages = []

        self._init_widgets()

    def _init_widgets(self):
        # contents
        contents = QListWidget()
        contents.setViewMode(QListView.ListMode)
        contents.setMovement(QListView.Static)
        # set the width to match the width of the content
        contents.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)

        def item_changed(item: QListWidgetItem):
            pageno: Page = item.data(1)
            pages.setCurrentIndex(pageno)

        contents.itemClicked.connect(item_changed)

        self._pages.append(Integration())
        self._pages.append(ThemeAndColors())
        self._pages.append(Style())

        pages = QStackedWidget()
        for idx, page in enumerate(self._pages):
            pages.addWidget(page)
            list_item = QListWidgetItem(page.NAME)
            list_item.setData(1, idx)
            contents.addItem(list_item)

        # buttons
        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Close | QDialogButtonBox.StandardButton.Ok)
        buttons.button(QDialogButtonBox.Ok).setText("Save")
        buttons.accepted.connect(self._on_ok_clicked)
        buttons.rejected.connect(self.close)

        # layout
        top_layout = QHBoxLayout()
        top_layout.addWidget(contents)
        top_layout.addWidget(pages)

        main_layout = QVBoxLayout()
        main_layout.addLayout(top_layout)
        main_layout.addWidget(buttons)

        self.setLayout(main_layout)

    def _on_ok_clicked(self):
        for page in self._pages:
            page.save_config()
        save_config()
        refresh_theme()  # Apply updates to theme
        self.close()



from PySide6.QtWidgets import QTabWidget

from .qconstraint_viewer import QConstraintViewer
from .qfiledesc_viewer import QFileDescriptorViewer
from .qmemory_viewer import QMemoryViewer
from .qregister_viewer import QRegisterViewer
from .qvextemps_viewer import QVEXTempsViewer


class StateInspector(QTabWidget):
    """
    Dispaly detail information for a selected state.
    """

    def __init__(self, workspace, state, parent=None):
        super().__init__(parent=parent)
        self.workspace = workspace
        self._state = state

        self._register_viewer: QRegisterViewer
        self._memory_viewer: QMemoryViewer
        self._vextemps_viewer: QVEXTempsViewer
        self._constraint_viewer: QConstraintViewer
        self._filedesc_viewer: QFileDescriptorViewer

        self._init_widgets()

    def _init_widgets(self):
        self._register_viewer = QRegisterViewer(self._state, self, self.workspace)
        self.addTab(self._register_viewer, "Registers")

        self._memory_viewer = QMemoryViewer(self._state, self, self.workspace)
        self.addTab(self._memory_viewer, "Memory")

        self._constraint_viewer = QConstraintViewer(self._state, self, self.workspace)
        self.addTab(self._constraint_viewer, "Constraints")

        self._filedesc_viewer = QFileDescriptorViewer(self._state, self, self.workspace)
        self.addTab(self._filedesc_viewer, "File Descriptors")

        self._vextemps_viewer = QVEXTempsViewer(self._state, self, self.workspace)
        self.addTab(self._vextemps_viewer, "Temps")
